import pytest
import json
import os

HOME_PATH = os.path.dirname(os.path.abspath(__file__))


def test_version():
    from cbinterface import __version__

    assert __version__ == "2.3.20"


def test_timezone_settings():
    from dateutil import tz
    from cbinterface.config import set_timezone, get_timezone

    set_timezone("GMT")
    assert "GMT" == os.environ["CBINTERFACE_TIMEZONE"]
    assert get_timezone() == tz.gettz("GMT")


def test_cbapi_environment_settings():
    from cbinterface.config import (
        set_default_cbapi_product,
        set_default_cbapi_profile,
        get_default_cbapi_product,
        get_default_cbapi_profile,
    )

    set_default_cbapi_product("buckets")
    assert get_default_cbapi_product() == "buckets"
    set_default_cbapi_product("more_buckets")
    set_default_cbapi_profile("stacks")
    assert f"{get_default_cbapi_product()}:{get_default_cbapi_profile()}" == "more_buckets:stacks"


def test_playbook_map():
    from cbinterface.config import get_playbook_map

    playbook_map = get_playbook_map()
    assert isinstance(playbook_map, dict)
    playbook = list(playbook_map.values())[0]
    assert list(playbook.keys()) == ["path", "name", "description"]


def test_playbook_build():
    from cbinterface.scripted_live_response import build_playbook_commands
    from cbinterface.commands import GetFile, PutFile, ExecuteCommand

    commands = build_playbook_commands(f"{HOME_PATH}/test_data/playbooking.around.ini")
    # order matters
    assert len(commands) == 3
    assert isinstance(commands[0], PutFile)
    assert isinstance(commands[1], ExecuteCommand)
    assert isinstance(commands[2], GetFile)


def test_remediation_script():
    from cbinterface.scripted_live_response import build_remediation_commands
    from cbinterface.commands import (
        GetFile,
        PutFile,
        ExecuteCommand,
        DeleteFile,
        KillProcessByID,
        KillProcessByName,
        DeleteRegistryKeyValue,
        DeleteRegistryKey,
    )

    cmds = build_remediation_commands(f"{HOME_PATH}/test_data/remediating.around.ini")
    assert isinstance(cmds, list)
    # for cmd in cmds:
    #    print(type(cmd))
    assert len(cmds) == 16
    assert isinstance(cmds[0], KillProcessByID)
    assert isinstance(cmds[1], KillProcessByName)
    assert isinstance(cmds[2], ExecuteCommand)
    assert isinstance(cmds[3], GetFile)
    assert isinstance(cmds[4], ExecuteCommand)
    assert isinstance(cmds[5], ExecuteCommand)
    assert isinstance(cmds[6], GetFile)
    assert isinstance(cmds[7], ExecuteCommand)
    assert isinstance(cmds[8], DeleteRegistryKeyValue)
    assert isinstance(cmds[9], DeleteFile)
    assert isinstance(cmds[10], ExecuteCommand)
    assert isinstance(cmds[11], GetFile)
    assert isinstance(cmds[12], ExecuteCommand)
    assert isinstance(cmds[13], ExecuteCommand)
    assert isinstance(cmds[14], GetFile)
    assert isinstance(cmds[15], ExecuteCommand)



import pytest
import json
import os

from io import StringIO
from contextlib import redirect_stdout

from cbapi.psc.threathunter import CbThreatHunterAPI, Process
from cbapi.errors import ObjectNotFoundError

from cbinterface.config import set_timezone

HOME_PATH = os.path.dirname(os.path.abspath(__file__))


@pytest.fixture(autouse=True)
def no_requests(monkeypatch):
    """Remove requests.sessions.Session.request for all tests."""
    monkeypatch.delattr("requests.sessions.Session.request")


# helper
def fake_cb_api(monkeypatch):
    """Fake CbThreatHunterAPI object."""
    return CbThreatHunterAPI(url="https://mcfeely.net", token="N/A", org_key="ork_gey", ssl_verify=False)


def test_fake_cb(monkeypatch):
    assert isinstance(fake_cb_api(monkeypatch), CbThreatHunterAPI)


# helper
def get_dummy_process_data():
    process_guid = "H8NDJUE1-02361dc7-000009d4-00000000-1d70b8a6f55bfa7"
    initial_data = {}
    with open(f"{HOME_PATH}/test_data/{process_guid}.json", "r") as fp:
        initial_data = json.load(fp)
    return initial_data


# helper
def load_dummy_process(monkeypatch):
    initial_data = get_dummy_process_data()
    cb = fake_cb_api(monkeypatch)
    process_info = initial_data["info"]
    process_info["_events"] = initial_data["events"]
    return Process(cb, initial_data["info"]["process_guid"], initial_data=process_info)


def test_dummy_process(monkeypatch):
    assert isinstance(load_dummy_process(monkeypatch), Process)


def test_make_device_query(monkeypatch):
    from cbapi.psc.devices_query import DeviceSearchQuery
    from cbinterface.psc.device import make_device_query

    def _count(self):
        return 0

    cb = fake_cb_api(monkeypatch)
    monkeypatch.setattr(DeviceSearchQuery, "_count", _count)
    assert isinstance(make_device_query(cb, "name:test"), DeviceSearchQuery)


def test_make_process_query(monkeypatch):
    from cbapi.psc.threathunter.models import AsyncProcessQuery
    from cbinterface.psc.query import make_process_query

    cb = fake_cb_api(monkeypatch)
    assert isinstance(make_process_query(cb, "process_name:loop.exe", raise_exceptions=False), AsyncProcessQuery)


def test_is_valid_process_query(monkeypatch):
    from cbapi.psc.threathunter.models import AsyncProcessQuery
    from cbinterface.psc.query import make_process_query
    from cbinterface.psc.query import is_valid_process_query

    def _get_object(url, query_parameters):
        assert url == "/api/investigate/v1/orgs/ork_gey/processes/search_validation"
        assert query_parameters["q"] == "process_name:loop.exe"
        return {"valid": True}

    cb = fake_cb_api(monkeypatch)
    monkeypatch.setattr(cb, "get_object", _get_object)
    query = make_process_query(cb, "process_name:loop.exe", raise_exceptions=False)
    assert is_valid_process_query(query) is True


def test_is_process_loaded(monkeypatch):
    from cbinterface.psc.process import is_process_loaded

    p = load_dummy_process(monkeypatch)
    assert is_process_loaded(p) is True


def test_process_to_dict(monkeypatch, mocker):
    from cbapi.psc.threathunter.models import Event
    from cbinterface.psc.process import process_to_dict

    data = get_dummy_process_data()
    p = load_dummy_process(monkeypatch)

    def _events(self):
        all_events = []
        # NOTE, could use this same events HACK to allow users to parse events from json as-if from the Cb PSC.
        for etype in p._events.keys():
            all_events.extend(data["events"][etype])
        return [Event(p._cb, initial_data=e) for e in all_events]

    monkeypatch.setattr(Process, "events", _events)
    mocker.patch("cbinterface.psc.process.print_ancestry", return_value=data["process_ancestry"])
    mocker.patch("cbinterface.psc.process.print_process_tree", return_value=data["process_tree"])
    process_dict = process_to_dict(p)
    assert isinstance(process_dict, dict)
    assert process_dict.keys() == data.keys()
    assert process_dict["events"] == data["events"]



# PYTHON_ARGCOMPLETE_OK

import os
import re

import time
import argparse
import argcomplete
import logging
import coloredlogs
import datetime
import json
import signal
import yaml

import cbapi.auth
from cbapi.psc.threathunter import CbThreatHunterAPI
from cbapi.response import CbResponseAPI
from cbapi.errors import ConnectionError, UnauthorizedError, ServerError

from cbinterface.helpers import is_uuid, clean_exit, input_with_timeout
from cbinterface.config import (
    set_timezone,
    save_configuration,
    get_default_cbapi_product,
    get_default_cbapi_profile,
    set_default_cbapi_profile,
    set_default_cbapi_product,
    get_playbook_map,
)

from cbinterface.response.cli import add_response_arguments_to_parser, execute_response_arguments
from cbinterface.psc.cli import add_psc_arguments_to_parser, execute_threathunter_arguments
from cbinterface.scripted_live_response import write_playbook_template, write_remediation_template

LOGGER = logging.getLogger("cbinterface.cli")

SUPPORTED_PRODUCTS = ["response", "psc"]


def load_configured_environments():
    """Load Carbon Black environments from config files."""
    # set custom attributes
    default_profile = cbapi.auth.default_profile
    default_profile["lr_token"] = None  # needed for psc

    configured_environments = {}
    for product in SUPPORTED_PRODUCTS:
        configured_environments[product] = []
        # FileCredentialStore loads `default_profile`
        for profile in cbapi.auth.FileCredentialStore(product).get_profiles():
            configured_environments[product].append(profile)

    return configured_environments


def main():
    """Main entry point for cbinterface."""

    # configure logging #
    logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(name)s - [%(levelname)s] %(message)s")
    coloredlogs.install(level="INFO", logger=logging.getLogger())

    # set clean exit signal
    signal.signal(signal.SIGINT, clean_exit)

    # load carbonblack environment profiles #
    configured_environments = load_configured_environments()
    environments = []
    # create human friendly options for the CLI
    for product, profiles in configured_environments.items():
        for profile in profiles:
            environments.append(f"{product}:{profile}")

    # chose the default environment
    default_product_name = get_default_cbapi_product()
    default_profile_name = get_default_cbapi_profile()
    default_environments = [env for env in environments if env.startswith(default_product_name)]
    default_environment = f"{default_product_name}:{default_profile_name}"
    default_environment = (
        default_environment if default_environments and default_environment in default_environments else environments[0]
    )

    parser = argparse.ArgumentParser(description="Interface to Carbon Black for IDR teams.")
    parser.add_argument("-d", "--debug", action="store_true", help="Turn on debug logging.")
    parser.add_argument(
        "-e",
        "--environment",
        action="store",
        choices=environments,
        default=default_environment,
        help=f"specify an environment to work with. Default={default_environment}",
    )
    parser.add_argument(
        "-sde",
        "--set-default-environment",
        action="store",
        choices=environments,
        help="configure your default Cb environment",
    )
    parser.add_argument(
        "-tz",
        "--time-zone",
        action="store",
        help='specify the timezone to override defaults. ex. "US/Eastern" or "Europe/Rome"',
    )
    parser.add_argument(
        "--set-default-timezone",
        action="store",
        help='configure your default timezone. ex. "US/Eastern" or "Europe/Rome"',
    )

    subparsers = parser.add_subparsers(dest="command")

    # query parser
    parser_query = subparsers.add_parser(
        "query", aliases=["pq", "q"], help="Execute a process search query. 'query -h' for more"
    )
    parser_query.add_argument("query", help="the process search query you'd like to execute")
    parser_query.add_argument(
        "-s",
        "--start-time",
        action="store",
        help="Start time of the process.  Format:'Y-m-d H:M:S' UTC",
    )
    parser_query.add_argument(
        "-e",
        "--last-time",
        action="store",
        help="Narrow to processes with start times BEFORE this end/last time. Format:'Y-m-d H:M:S' UTC",
    )
    parser_query.add_argument(
        "-nw",
        "--no-warnings",
        action="store_true",
        default=False,
        help="Don't warn before printing large query results",
    )
    parser_query.add_argument(
        "-ad",
        "--all-details",
        action="store_true",
        default=False,
        help="Print all available process info (all fields).",
    )
    parser_query.add_argument(
        "--facets", action="store_true", default=None, help="Retrieve statistical facets for this query."
    )

    # process inspection/investigation parser
    parser_inspect = subparsers.add_parser(
        "investigate", aliases=["proc", "i"], help="Investigate process events and metadata."
    )
    parser_inspect.add_argument(
        "process_guid_options", help="the process GUID/segment to inspect. Segment is optional."
    )
    parser_inspect.add_argument(
        "-i", "--proc-info", dest="inspect_proc_info", action="store_true", help="show binary and process information"
    )
    parser_inspect.add_argument(
        "-w",
        "--walk-tree",
        dest="walk_and_inspect_tree",
        action="store_true",
        help="Recursively walk, print, and inspect the process tree. Specified arguments (ex. filemods) applied at every process in tree. WARNING: can pull large datasets.",
    )
    parser_inspect.add_argument(
        "-t",
        "--process-tree",
        dest="inspect_process_tree",
        action="store_true",
        help="print the process tree with this process as the root.",
    )
    parser_inspect.add_argument(
        "-a",
        "--process-ancestry",
        dest="inspect_process_ancestry",
        action="store_true",
        help="print the the process ancestry",
    )
    parser_inspect.add_argument(
        "-c",
        "--show-children",
        dest="inspect_children",
        action="store_true",
        help="only print process children event details",
    )
    parser_inspect.add_argument(
        "-nc", "--netconns", dest="inspect_netconns", action="store_true", help="print network connections"
    )
    parser_inspect.add_argument(
        "-fm", "--filemods", dest="inspect_filemods", action="store_true", help="print file modifications"
    )
    parser_inspect.add_argument(
        "-rm", "--regmods", dest="inspect_regmods", action="store_true", help="print registry modifications"
    )
    parser_inspect.add_argument(
        "-ml", "--modloads", dest="inspect_modloads", action="store_true", help="print modloads"
    )
    parser_inspect.add_argument(
        "-sl", "--scriptloads", dest="inspect_scriptloads", action="store_true", help="print scriptloads (PSC)"
    )
    parser_inspect.add_argument(
        "-cp", "--crossprocs", dest="inspect_crossprocs", action="store_true", help="print crossprocs"
    )
    parser_inspect.add_argument(
        "-rpe",
        "--raw-print-events",
        action="store_true",
        default=False,
        help="do not format Cb events onto a single line. Print them the way Cb does by default.",
    )
    # parser_inspect.add_argument('-warn', '--give-warnings', action='store_true', default=False,
    #                         help="Warn before printing large datasets/results")
    parser_inspect.add_argument(
        "--json", action="store_true", help="Combine all results into json document and print the result."
    )
    parser_inspect.add_argument(
        "--segment-limit",
        action="store",
        type=int,
        default=None,
        help="stop processing events into json after this many process segments",
    )

    # live response parser
    parser_lr = subparsers.add_parser(
        "live-response", aliases=["lr"], help="Perform live response actions on a device/sensor."
    )
    parser_lr.add_argument("name_or_id", help="the hostname or sensor/device id to go live with.")
    parser_lr.add_argument(
        "-e", "--execute-command", action="store", help="Execute this command on the sensor. NOTE: waits for output."
    )
    parser_lr.add_argument("-cr", "--create-regkey", action="store", help="Create this regkey.")
    parser_lr.add_argument("-sr", "--set-regkey-value", action="append", help="Set this regkey value.")
    if configured_environments["response"]:
        parser_lr.add_argument(
            "-i",
            "--sensor-isolation-toggle",
            action="store_true",
            help="Sensor hostname/ID to isolation/unisolate (on/off). (CB Response)",
        )
    if configured_environments["psc"]:
        parser_lr.add_argument(
            "-q",
            "--quarantine",
            action="store_true",
            default=False,
            help="Quarantine the devices returned by the query. (PSC)",
        )
        parser_lr.add_argument(
            "-uq",
            "--un_quarantine",
            action="store_true",
            default=False,
            help="UN-Quarantine the devices returned by the query. (PSC)",
        )

    # live response subparser
    lr_subparsers = parser_lr.add_subparsers(dest="live_response_command")

    # live response put file parser
    parser_put_file = lr_subparsers.add_parser("put", help="Put a file on the device/sensor.")
    parser_put_file.add_argument("local_filepath", action="store", help="Path to the file.")
    parser_put_file.add_argument("sensor_write_filepath", action="store", help="Path to write the file on the sensor.")

    # live response playbook parser
    parser_playbook = lr_subparsers.add_parser(
        "playbook", aliases=["pb", "play"], help="Execute a live response playbook script."
    )
    parser_playbook.add_argument(
        "-f", "--playbook-configpath", action="store", help="Path to a playbook config file to execute."
    )
    playbook_map = get_playbook_map()
    playbook_names = [p["name"] for _, p in playbook_map.items()]
    parser_playbook.add_argument(
        "-p",
        "--playbook-name",
        action="store",
        choices=playbook_names,
        help="The name of a configured playbook to execute.",
    )
    parser_playbook.add_argument("-l", "--list-playbooks", action="store_true", help="List configured playbooks.")
    parser_playbook.add_argument(
        "--write-template", action="store_true", help="write a playbook template file to use as example."
    )

    # live response collect parser
    parser_collect = lr_subparsers.add_parser("collect", help="Collect artifacts from hosts.")
    parser_collect.add_argument(
        "-i", "--sensor-info", dest="sensor_info", action="store_true", help="print default sensor information"
    )
    parser_collect.add_argument("-p", "--process-list", action="store_true", help="show processes running on sensor")
    parser_collect.add_argument("-f", "--file", action="store", help="collect file at this path on sensor")
    parser_collect.add_argument(
        "-lr", "--regkeypath", action="store", help="List all registry values from the specified registry key."
    )
    parser_collect.add_argument(
        "-r", "--regkeyvalue", action="store", help="Returns the associated value of the specified registry key."
    )
    parser_collect.add_argument(
        "-ld", "--list-directory", action="store", help="List the contents of a directory on the sensor."
    )
    parser_collect.add_argument(
        "-wd", "--walk-directory", action="store", help="List the contents of a directory on the sensor."
    )
    parser_collect.add_argument("--drives", action="store_true", help="Get logical drives on this sensor.")
    parser_collect.add_argument(
        "--memdump", action="store_true", help="Use Cb to dump sensor memory and collect the memdump."
    )

    # live response remediation parser
    parser_remediate = lr_subparsers.add_parser(
        "remediate", help="Perform remdiation (delete/kill) actions on device/sensor."
    )
    parser_remediate.add_argument(
        "-f", "--delete-file-path", action="store", help="delete the file at this path on the sensor"
    )
    parser_remediate.add_argument(
        "-kpname", "--kill-process-name", action="store", help="kill all processes with this name"
    )
    parser_remediate.add_argument("-kpid", "--kill-process-id", action="store", help="kill the process with this ID")
    parser_remediate.add_argument("-drv", "--delete-regkeyvalue", action="store", help="Delete the regkey value.")
    parser_remediate.add_argument(
        "--delete-entire-regkey", action="store", help="Delete the registry key and all values. BE CAREFUL."
    )
    parser_remediate.add_argument("-rs", "--remediation-script", action="store", help="Path to a remediaiton script.")
    parser_remediate.add_argument("--write-template", action="store_true", help="write a remediation template.")

    # session parser - NOTE: functionality is limited on the PSC side, and it's specifically annoying that
    # we can not get a list of active psc lr sessions... or at least I haven't figure out how to do that.
    parser_session = subparsers.add_parser("session", help="Interact with Cb live response server sessions.")
    if configured_environments["response"]:
        parser_session.add_argument(
            "-lss",
            "--list-sensor-sessions",
            action="store",
            help="list all CbLR sessions associated to this sensor ID (Response only).",
        )
    parser_session.add_argument(
        "-gsc", "--get-session-command-list", action="store", help="list commands associated to this session"
    )
    if configured_environments["response"]:
        parser_session.add_argument(
            "-a", "--list-all-sessions", action="store_true", help="list all CbLR sessions (Response only)."
        )
    parser_session.add_argument("-g", "--get-session", action="store", help="get live response session by id.")
    parser_session.add_argument("-c", "--close-session", action="store", help="close live response session by id.")
    parser_session.add_argument(
        "-gcr", "--get-command-result", action="store", help="get any results for this command."
    )
    parser_session.add_argument(
        "-f", "--get-file-content", action="store", help="byte stream any file content to stdout. (use a pipe)"
    )

    # enumeration parser
    parser_enumeration = subparsers.add_parser(
        "enumerate", aliases=["e"], help="Data enumerations for answering common questions."
    )
    parser_enumeration.add_argument(
        "-lh",
        "--logon-history",
        action="store",
        help="Given process username or device name, roughly enumerate logon history (Windows OS).",
    )

    # only add independent product args if product is a configured option
    if configured_environments["response"]:
        add_response_arguments_to_parser(subparsers)
    if configured_environments["psc"]:
        add_psc_arguments_to_parser(subparsers)

    argcomplete.autocomplete(parser)
    args = parser.parse_args()

    if args.debug:
        logging.getLogger("urllib3.connectionpool").setLevel(logging.INFO)
        coloredlogs.install(level="DEBUG", logger=logging.getLogger())

    if args.time_zone:
        set_timezone(args.time_zone)

    if args.set_default_timezone:
        set_timezone(args.set_default_timezone)
        save_configuration()

    if args.set_default_environment:
        product, profile = args.set_default_environment.split(":", 1)
        set_default_cbapi_product(product)
        set_default_cbapi_profile(profile)
        save_configuration()

    # Functionality that doesn't require a Cb connection.
    if args.command and (args.command.lower() == "lr" or args.command.lower().startswith("live")):
        if args.live_response_command and (
            args.live_response_command.startswith("play") or args.live_response_command == "pb"
        ):
            if args.list_playbooks:
                print(f"\nConfigured Playbooks:")
                for pb_key, pb_metadata in playbook_map.items():
                    print(f"\t{pb_metadata['name']} : {pb_metadata['description']}")
                print()
                return True
            if args.write_template:
                template_path = write_playbook_template()
                if os.path.exists(template_path):
                    LOGGER.info(f" + wrote {template_path}")
                return True
        if args.live_response_command and args.live_response_command.startswith("r"):
            if args.write_template:
                template_path = write_remediation_template()
                if os.path.exists(template_path):
                    LOGGER.info(f" + wrote {template_path}")
                return True

    # Connect and execute
    product, profile = args.environment.split(":", 1)
    LOGGER.debug(f"using '{profile}' profile via the configured '{product}' product.")
    try:
        if product == "response":
            cb = CbResponseAPI(profile=profile)
            execute_response_arguments(cb, args)

        elif product == "psc":
            cb = CbThreatHunterAPI(profile=profile)
            execute_threathunter_arguments(cb, args)
    except ConnectionError as e:
        LOGGER.critical(f"Couldn't connect to {product} {profile}: {e}")
    except UnauthorizedError as e:
        LOGGER.critical(f"{e}")
    except ServerError as e:
        LOGGER.critical(f"CB ServerError 😒 (try again) : {e}")
    except TimeoutError as e:
        LOGGER.critical(f"TimeoutError waiting for CB server 🙄 (try again) : {e}")



"""Functions that work with Carbon Black Sensors.
"""

import datetime
import logging

from typing import Union

from cbapi.response import CbResponseAPI, Sensor
from cbapi.response.models import SensorQuery

from cbinterface.helpers import as_configured_timezone

LOGGER = logging.getLogger("cbinterface.response.sensor")


def is_sensor_online(s: SensorQuery) -> bool:
    """Return True if the sensor is online."""
    s.refresh()
    LOGGER.debug(f"sensor status: {s.status}")
    if s.status.lower() == "online":
        return True
    return False


def make_sensor_query(cb: CbResponseAPI, sensor_query: str) -> SensorQuery:
    """Construct a SensorQuery object."""
    try:
        if ":" not in sensor_query:
            LOGGER.warning("No field specification passed. Fields: ip, hostname, groupid")
            LOGGER.info(f"Making assumption and updating query to: 'hostname:{sensor_query}'")
            sensor_query = f"hostname:{sensor_query}"
        sensors = cb.select(Sensor).where(sensor_query)
    except ValueError as e:
        LOGGER.error(f"{e}")
        return False
    LOGGER.info(f"got {len(sensors)} sensor results.")
    return sensors


def find_sensor_by_hostname(cb: CbResponseAPI, hostname: str) -> Sensor:
    """Find a Sensor by hostname."""
    sensors = make_sensor_query(cb, f"hostname:{hostname}")
    if len(sensors) == 1:
        return sensors[0]
    elif len(sensors) > 1:
        LOGGER.warning(f"{len(sensors)} sensors with hostname={hostname}")
        return None
    return None


def sensor_info(sensor: Sensor):
    from cbinterface.helpers import utc_offset_to_potential_tz_names

    text = "\n"
    text += f"Sensor object - {sensor.webui_link}\n"
    text += "-------------------------------------------------------------------------------\n"
    text += f"\tcb_build_version_string: {sensor.build_version_string}\n"
    text += f"\tcomputer_sid: {sensor.computer_sid}\n"
    text += f"\tcomputer_dns_name: {sensor.computer_dns_name}\n"
    text += f"\tcomputer_name: {sensor.computer_name}\n"
    text += f"\tos_environment_display_string: {sensor.os_environment_display_string}\n"
    text += f"\tphysical_memory_size: {sensor.physical_memory_size}\n"
    text += f"\tsystemvolume_free_size: {sensor.systemvolume_free_size}\n"
    text += f"\tsystemvolume_total_size: {sensor.systemvolume_total_size}\n"
    text += "\n"
    text += f"\tstatus: {sensor.status}\n"
    text += f"\tis_isolating: {sensor.is_isolating}\n"
    text += f"\tsensor_id: {sensor.id}\n"
    potential_tz_names = utc_offset_to_potential_tz_names(sensor.last_checkin_time.tzinfo._offset)
    potential_regions = f"{len(potential_tz_names)} potential regions."
    if len(potential_tz_names) == 1:
        potential_regions = potential_tz_names[0]
    text += f"\tsensor_timezone: GMT{sensor.last_checkin_time.strftime('%z')} - {potential_regions}\n"
    text += f"\tlast_checkin_time: {as_configured_timezone(sensor.last_checkin_time)}\n"
    text += f"\tnext_checkin_time: {as_configured_timezone(sensor.next_checkin_time)}\n"
    text += f"\tsensor_health_message: {sensor.sensor_health_message}\n"
    text += f"\tsensor_health_status: {sensor.sensor_health_status}\n"
    text += "\tnetwork_interfaces:\n"
    for ni in sensor.network_interfaces:
        text += f"\t\t{ni}\n"
    return text



"""Everthing ThreatHunter process related.
"""

import json
import logging

from io import StringIO
from datetime import datetime
from contextlib import redirect_stdout
from typing import Dict, Union

from cbapi.psc.threathunter import CbThreatHunterAPI, Process
from cbapi.errors import ObjectNotFoundError

from cbinterface.helpers import as_configured_timezone, get_os_independent_filepath

LOGGER = logging.getLogger("cbinterface.psc.process")


def load_process(p: Process) -> Process:
    """Load any process meta-data that exists or return None."""
    try:
        return Process.new_object(p._cb, p.summary._info["process"])
    except RecursionError:
        LOGGER.warning(f"RecursionError occurred loading process details.. loading incomplete details.")
        url = f"/api/investigate/v1/orgs/{p._cb.credentials.org_key}/processes/summary"
        summary = p._cb.get_object(url, query_parameters={"process_guid": p.process_guid})
        return Process.new_object(p._cb, summary["process"])
    except ObjectNotFoundError:
        LOGGER.debug(f"Process data does not exist for GUID={p.get('process_guid')}")
        return None


def select_process(cb: CbThreatHunterAPI, process_guid: str):
    """Select and load a Cb processes with a single API call.

    MUCH faster than using the suggested method.

    Args:
        cb: a CbThreatHunterAPI object
        process_guid: a Cb PSC GUID in the form of a string.
    Returns:
        A threathunter.Process or None
    """
    # Create empty Process. `force_init` is not supported (underlying PSC model is unrefreshable)
    # Use the process object (see load_process) to load a new process object containing the completed process meta-data.
    return load_process(Process(cb, process_guid))


def is_process_loaded(p: Process) -> bool:
    """Return True if the process has data."""
    if len(p._info.keys()) < 3:
        return False
    return True


def process_to_dict(p: Process, max_events: int = None) -> Dict:
    """Convert process and it's events to a dictionary.

    Args:
      p: Carbon Black Cloud Process
      max_events: a limit on the amount of process events to add.
    Returns:
      Dictionary representation of the process
    """
    # TODO: add start & end args so process events during a specific
    #  time window can be pulled. This could be added for several functions on this page.
    process = {}
    if not is_process_loaded(p):
        p = load_process(p)
    process["info"] = p._info
    process["events"] = {}
    event_count = 0
    for event in p.events():
        if event.event_type not in process["events"]:
            process["events"][event.event_type] = []
        process["events"][event.event_type].append(event._info)
        event_count += 1
        if max_events is not None:
            assert isinstance(max_events, int)
            if event_count >= max_events:
                LOGGER.info(f"max event limit of {max_events} reached.")
                break

    process["process_ancestry"] = StringIO()
    with redirect_stdout(process["process_ancestry"]):
        print_ancestry(p)
    process["process_ancestry"] = process["process_ancestry"].getvalue()

    process["process_tree"] = StringIO()
    with redirect_stdout(process["process_tree"]):
        print_process_tree(p)
    process["process_tree"] = process["process_tree"].getvalue()

    return process


def print_process_info(proc: Process, return_string: bool = False, raw_print=False, header=True):
    """Analyst friendly custom process data format.

    Args:
        proc: CbTH Process (fully initialized)
        return_string: return string if True, else print it to stdout.
    Returns: string or None
    """

    if not is_process_loaded(proc):
        proc = load_process(proc)

    txt = ""
    if header:
        txt += "------ INFO ------\n"
    if raw_print:
        txt = str(proc)
    else:
        txt += f"  Process GUID: {proc.get('process_guid')}\n"
        process_name = get_os_independent_filepath(proc.get("process_name", "None")).name
        txt += f"  Process Name: {process_name}\n"
        process_pid = [str(_) for _ in proc.get("process_pid", [])]
        txt += f"  Process PID: {', '.join(process_pid)}\n"
        txt += f"  Process MD5: {proc.get('process_md5')}\n"
        txt += f"  Process SHA256: {proc.get('process_sha256')}\n"
        txt += f"  Process Path: {proc.get('process_name')}\n"
        txt += f"  Process Terminated: {proc.get('process_terminated')}\n"
        txt += f"  Start Time: {as_configured_timezone(proc.get('process_start_time', ''))}\n"
        process_command_line = (
            proc.process_cmdline[0]
            if proc.get("process_cmdline") and len(proc.process_cmdline) == 1
            else proc.get("process_cmdline")
        )
        txt += f"  Command Line: {process_command_line}\n"
        txt += f"  Process Reputation: {proc.get('process_reputation')}\n"
        txt += f"  Parent Name: {proc.get('parent_name')}\n"
        txt += f"  Parent GUID: {proc.get('parent_guid')}\n"
        parent_sha256 = next((hsh for hsh in proc.get("parent_hash", []) if len(hsh) == 64), None)
        txt += f"  Parent SHA256: {parent_sha256}\n"
        txt += f"  Username: {proc.get('process_username')}\n"
        txt += f"  Device ID: {proc.get('device_id')}\n"
        txt += f"  Device Name: {proc.get('device_name')}\n"
        txt += f"  Device OS: {proc.get('device_os')}\n"
        txt += f"  External IP: {proc.get('device_external_ip')}\n"
        txt += f"  Internal IP: {proc.get('device_internal_ip')}\n"
    if return_string:
        return txt
    txt += "\n"
    print(txt)


def print_ancestry(p: Process, max_depth=0, depth=0):
    """Print the process ancestry for this process."""

    if max_depth and depth > max_depth:
        return

    if depth == 0:
        print("\n------ Process Ancestry ------")
        print()

    if not is_process_loaded(p):
        p = load_process(p)

    start_time = as_configured_timezone(p.get("process_start_time", ""))
    command_line = (
        p.process_cmdline[0] if p.get("process_cmdline") and len(p.process_cmdline) == 1 else p.get("process_cmdline")
    )
    print(f"{'  '*(depth + 1)}{start_time}: {command_line} | {p.process_guid}")

    if p.get("parent_guid"):
        parent = select_process(p._cb, p.parent_guid)
        if parent:
            print_ancestry(parent, max_depth=max_depth, depth=depth + 1)


def print_process_tree(p: Process, max_depth=0, depth=0):
    """Print the process tree."""

    if max_depth and depth > max_depth:
        return

    if depth == 0:
        if not is_process_loaded(p):

            p = load_process(p)
        print("\n------ Process Execution Tree ------")
        print()

    command_line = (
        p.process_cmdline[0] if p.get("process_cmdline") and len(p.process_cmdline) == 1 else p.get("process_cmdline")
    )
    print(f"  {'  '*(depth+1)}{command_line}  | {p.process_guid}")

    try:
        for child in p.children:
            print_process_tree(child, max_depth=max_depth, depth=depth + 1)
    except RecursionError:
        LOGGER.warning(f"hit RecursionError walking process tree.. stopping here")
        print(" [!] reached recursion limit walking process tree ...")


def get_events_by_type(p: Union[Process, Dict], event_type: str):
    """Return process events by event type.

    One of filemod, netconn, regmod, modload, crossproc, childproc, scriptload,
        fileless_scriptload
    """
    if isinstance(p, dict):
        # processing from json file
        assert "events" in p
        return p["events"].get(event_type, [])

    # we only load the process here to check for the presence of event_types
    # only make event api calls for processes that report having those events
    """# XXX - UPDATE - NOTE: These processes LIE; sometimes procs say they don't have events when they do.
    if not is_process_loaded(p):
        p = load_process(p)
    event_count = p.get(f"{event_type}_count")
    if event_count is not None and event_count == 0:
        LOGGER.debug(f"processes reports not having any events of type: {event_type}")
        return []
    """
    try:
        return p.events(event_type=event_type)
    except Exception as e:
        LOGGER.error(f"failed to get events: {e}")
        return []


def print_filemods(p: Union[Process, Dict], raw_print=False, return_string: bool = False, **kwargs):
    """Print file modifications.

    one or more of ACTION_INVALID, ACTION_FILE_CREATE, ACTION_FILE_WRITE, ACTION_FILE_DELETE, ACTION_FILE_LAST_WRITE, ACTION_FILE_MOD_OPEN,
     ACTION_FILE_RENAME, ACTION_FILE_UNDELETE, ACTION_FILE_TRUNCATE, ACTION_FILE_OPEN_READ, ACTION_FILE_OPEN_WRITE, ACTION_FILE_OPEN_DELETE,
     ACTION_FILE_OPEN_EXECUTE, ACTION_FILE_READ
    """

    txt = "------ FILEMODS ------\n"
    for fm in get_events_by_type(p, "filemod"):
        if raw_print:
            txt += str(fm)
            continue
        _action_summary = [action[len("ACTION_") :] for action in fm.get("filemod_action", [])]
        _edge_actions = [action for action in _action_summary if not action.startswith("FILE")]
        _fm_action_summary = [action[len("FILE_") :] for action in _action_summary if action.startswith("FILE")]
        _fm_action_summary.extend(_edge_actions)
        action_summary = ",".join(_fm_action_summary)
        fm_sha256 = f" , sha256:{fm.get('filemod_sha256')}" if fm.get("filemod_sha256") else ""
        txt += f" @{as_configured_timezone(fm.get('event_timestamp'))}: |{action_summary}| {fm.get('filemod_name')}{fm_sha256}\n"
    txt += "\n"
    if return_string:
        return txt
    print(txt)


def print_netconns(p: Union[Process, Dict], raw_print=False, return_string: bool = False):
    """Print network connection events.

    action is one or more of: ACTION_CONNECTION_CREATE, ACTION_CONNECTION_CLOSE,
     ACTION_CONNECTION_ESTABLISHED, ACTION_CONNECTION_CREATE_FAILED, ACTION_CONNECTION_LISTEN
    """
    import socket, struct

    txt = "------ NETCONNS ------\n"
    for nc in get_events_by_type(p, "netconn"):
        if raw_print:
            txt += str(nc)
            continue
        action = (
            nc.get("netconn_action", "")[len("ACTION_CONNECTION_") :]
            if nc.get("netconn_action", "").startswith("ACTION_CONNECTION_")
            else nc.get("netconn_action")
        )
        if action == "CREATE":
            # same behavior as PSC GUI
            action = "ESTABLISHED"
        action = action.capitalize()
        protocol = nc.get("netconn_protocol", "")
        if protocol.startswith("PROTO_"):
            protocol = protocol[len("PROTO_") :]
        direction = "inbound" if nc.get("netconn_inbound") else "outbound"

        # NOTE: Cb stores ipv4 as integers and returns them as such. Their documentation is lacking but
        # research suggests they follow https://tools.ietf.org/html/rfc1700 and big-endian int can be assumed.

        local_ipv4 = nc.get("netconn_local_ipv4", "")
        if local_ipv4:
            local_ipv4 = socket.inet_ntoa(struct.pack("!i", local_ipv4))
        local_ipv6 = nc.get("netconn_local_ipv6", "")
        if local_ipv6:
            # TODO: insert a colon character between every four alphanumeric characters
            local_ipv6 = f"ipv6({local_ipv6})"
        local = f"from {local_ipv4}{local_ipv6}:{nc.get('netconn_local_port')}"

        remote_ipv4 = nc.get("netconn_remote_ipv4", "")
        if remote_ipv4:
            remote_ipv4 = socket.inet_ntoa(struct.pack("!i", remote_ipv4))
        remote_ipv6 = nc.get("netconn_remote_ipv6", "")
        if remote_ipv6:
            # TODO: insert a colon character between every four alphanumeric characters
            remote_ipv6 = f"ipv6({remote_ipv6})"
        remote = f"to {remote_ipv4}{remote_ipv6}:{nc.get('netconn_remote_port')}"

        domain = f"(domain={nc.get('netconn_domain')})" if nc.get("netconn_domain") else ""
        txt += f" @{as_configured_timezone(nc.get('event_timestamp'))}: {action} {direction} {protocol} {local} {remote} {domain}\n"
    txt += "\n"
    if return_string:
        return txt
    print(txt)


def print_regmods(p: Union[Process, Dict], raw_print=False, return_string: bool = False):
    """Print registry modifications.

    Actions: ACTION_INVALID, ACTION_CREATE_KEY, ACTION_WRITE_VALUE, ACTION_DELETE_KEY, ACTION_DELETE_VALUE,
        ACTION_RENAME_KEY, ACTION_RESTORE_KEY, ACTION_REPLACE_KEY, ACTION_SET_SECURITY
    """
    txt = "------ REGMODS ------\n"
    for rm in get_events_by_type(p, "regmod"):
        if raw_print:
            txt += str(rm)
            continue
        actions = []
        for a in rm.get("regmod_action"):
            if a.startswith("ACTION_"):
                actions.append(a[len("ACTION_") :])
            else:
                actions.append(a)
        # how could the list they return ever be greater than 1 with un-corrupted data?
        action = actions[0]  # if len(actions) > 1 else actions
        if "CREATE" in action:
            action = "Created"
        elif action == "WRITE_VALUE":
            action = "Modified"
        txt += f" @{as_configured_timezone(rm.get('event_timestamp'))}: {action}: {rm.get('regmod_name')}\n"
    txt += "\n"
    if return_string:
        return txt
    print(txt)


def print_scriptloads(p: Union[Process, Dict], raw_print=False, return_string: bool = False):
    """Print scriptloads and fileless scriptloads."""
    txt = "------ SCRIPTLOADS ------\n"
    for sl in get_events_by_type(p, "scriptload"):
        if raw_print:
            txt += str(sl)
            continue

        pub_state = ",".join([state[len("FILE_SIGNATURE_") :] for state in sl.get("scriptload_publisher_state", [])])
        txt += f" @{as_configured_timezone(sl.get('event_timestamp'))}: {sl.get('scriptload_name')} , sha256={sl.get('scriptload_sha256')} - {pub_state}\n"

    txt += "\n"
    fileless_sl_events = get_events_by_type(p, "fileless_scriptload")
    if fileless_sl_events:
        txt += "------ FILELESS SCRIPTLOADS ------\n"
    for fsl in fileless_sl_events:
        if raw_print:
            txt += str(fsl)
            continue

        txt += f" @{as_configured_timezone(fsl.get('event_timestamp'))}: {fsl.get('fileless_scriptload_cmdline')}\n"
    txt += "\n"
    if return_string:
        return txt
    print(txt)


def print_modloads(p: Union[Process, Dict], raw_print=False, return_string: bool = False):
    """Print modual/library loads."""

    txt = "------ MODLOADS ------\n"
    for ml in get_events_by_type(p, "modload"):
        if raw_print:
            txt += str(ml)
            continue
        # "for now can only be: ACTION_LOAD_MODULE"
        # action = "Loaded" if ml.modload_action == 'ACTION_LOAD_MODULE' else ml.modload_action
        ml_pub_state_summary = (
            "_".join([state[len("FILE_SIGNATURE_STATE") + 1 :] for state in ml.get("modload_publisher_state", [])])
            or ""
        )
        txt += f" @{as_configured_timezone(ml.get('event_timestamp'))}: {ml.get('modload_name')} , md5:{ml.get('modload_md5')} - {ml.get('modload_publisher')}: {ml_pub_state_summary}\n"
    txt += "\n"
    if return_string:
        return txt
    print(txt)


def print_crossprocs(p: Union[Process, Dict], raw_print=False, return_string: bool = False):
    """Print Cross Process activity.

    Actions: ACTION_DUP_PROCESS_HANDLE, ACTION_OPEN_THREAD_HANDLE, ACTION_DUP_THREAD_HANDLE,
        ACTION_CREATE_REMOTE_THREAD, ACTION_API_CALL
    """
    txt = "------ CROSSPROCS ------\n"
    for cp in get_events_by_type(p, "crossproc"):
        if raw_print:
            txt += str(cp)
            continue
        actions = [a[len("ACTION_") :] for a in cp.get("crossproc_action", [])]
        if len(actions) == 1:
            actions = actions[0]
        else:
            actions = ",".join(actions)
        inverse_target = "from" if cp.get("crossproc_target") is True else "to"
        direction = "<-" if cp.get("crossproc_target") is True else "->"
        proc_guid_direction = f"{cp.get('process_guid')} {direction} {cp.get('crossproc_process_guid')}"
        txt += f" @{as_configured_timezone(cp.get('event_timestamp'))}: {actions} {inverse_target} {cp.get('crossproc_name')} ({cp.get('crossproc_sha256')}) | {proc_guid_direction}\n"
    txt += "\n"
    if return_string:
        return txt
    print(txt)


def print_childprocs(p: Union[Process, Dict], raw_print=False, return_string: bool = False):
    """Print child process events."""
    txt = "------ CHILDPROCS ------\n"
    for cp in get_events_by_type(p, "childproc"):
        if raw_print:
            txt += str(cp)
            continue
        txt += f" @{as_configured_timezone(cp.get('event_timestamp'))}: {cp.get('childproc_cmdline')}  - {cp.get('childproc_process_guid')}"
    txt += "\n"
    if return_string:
        return txt
    print(txt)


def inspect_process_tree(
    proc: Process,
    info=False,
    filemods=False,
    netconns=False,
    regmods=False,
    modloads=False,
    crossprocs=False,
    children=False,
    scriptloads=False,
    max_depth=0,
    depth=0,
    **kwargs,
):
    """Walk down the execution chain and print inspection points."""
    if max_depth and depth > max_depth:
        return

    if depth == 0:
        print_ancestry(proc)
        print_process_tree(proc)

    process_name = get_os_independent_filepath(proc.get("process_name", "None")).name
    print(f"\n+ {process_name} - {proc.process_guid}")
    if info:
        print_process_info(proc, **kwargs)
    if filemods:
        print_filemods(proc, **kwargs)
    if netconns:
        print_netconns(proc, **kwargs)
    if regmods:
        print_regmods(proc, **kwargs)
    if modloads:
        print_modloads(proc, **kwargs)
    if crossprocs:
        print_crossprocs(proc, **kwargs)
    if children:
        print_childprocs(proc, **kwargs)
    if scriptloads:
        print_scriptloads(proc, **kwargs)

    for child in proc.children:
        try:
            inspect_process_tree(
                child,
                info=info,
                filemods=filemods,
                netconns=netconns,
                regmods=regmods,
                modloads=modloads,
                crossprocs=crossprocs,
                children=children,
                scriptloads=scriptloads,
                max_depth=max_depth,
                depth=depth + 1,
                **kwargs,
            )
        except RecursionError:
            LOGGER.warning(f"hit RecursionError inspecting process tree.")
            break



"""Functions that work with query related Carbon Black APIs.
"""

import datetime
import logging
from dateutil import tz

from typing import Union

from cbapi.response import CbResponseAPI, Process
from cbapi.response.models import ProcessQuery

LOGGER = logging.getLogger("cbinterface.response.query")


def make_process_query(
    cb: CbResponseAPI,
    query: str,
    start_time: datetime.datetime = None,
    last_time: datetime.datetime = None,
    raise_exceptions=True,
) -> ProcessQuery:
    """Query the CbResponse environment and interface results.

    Args:
        cb: A CbResponseAPI object to use
        query: The correctly formated query
        start_time: Set the minimum last update time (relative to server) for this query.
        last_time: Set the maximum last update time (relative to server) for this query.
        XXX no_warnings: Do not warn before printing large query result sets.
    Returns: cbapi.response.models.ProcessQuery or empty list.
    """

    processes = []
    LOGGER.debug(f"buiding query: {query} between '{start_time}' and '{last_time}'")
    try:
        processes = cb.select(Process).where(query).group_by("id")
        processes = processes.min_last_server_update(start_time) if start_time else processes
        processes = processes.max_last_server_update(last_time) if last_time else processes
        LOGGER.info(f"got {len(processes)} process results grouped by id.")
    except Exception as e:
        if raise_exceptions:
            raise (e)
        LOGGER.error(f"problem querying carbonblack with '{query}' : {e}")

    return processes


def print_facet_histogram(facet_dict):
    """Print facets"""
    print("\n------------------------- FACET HISTOGRAMS -------------------------")
    for field_name, facets in facet_dict.items():
        if any([key for key in ["name", "percent", "ratio", "value"] if key not in facets[0].keys()]):
            continue
        # total_results = sum([entry['value'] for entry in facets])
        # longest_process_name = len(max(process_names, key=len))
        print(f"\n\t\t\t{field_name} results: {len(facets)}")  # .format(total_results))
        print("\t\t\t--------------------------")
        for entry in facets:
            print(
                "%50s: %5s %5s%% %s"
                % (entry["name"][:45], entry["value"], entry["ratio"], "\u25A0" * (int(entry["percent"] / 2)))
            )
        print()
    return



import sys
import re
import argparse
import datetime
import logging
import json
import time

from cbapi.response import CbResponseAPI, Process, Sensor
from cbapi.errors import ObjectNotFoundError

from cbinterface.helpers import is_uuid, clean_exit, input_with_timeout
from cbinterface.response.query import make_process_query, print_facet_histogram
from cbinterface.response.sensor import is_sensor_online, find_sensor_by_hostname, make_sensor_query, sensor_info
from cbinterface.response.watchlists import get_all_watchlists, query_watchlists, these_watchlists_to_list_dict
from cbinterface.response.process import (
    process_to_dict,
    inspect_process_tree,
    print_process_info,
    print_ancestry,
    print_process_tree,
    print_filemods,
    print_netconns,
    print_regmods,
    print_modloads,
    print_crossprocs,
    print_childprocs,
)
from cbinterface.response.sessions import (
    CustomLiveResponseSessionManager,
    get_session_by_id,
    sensor_live_response_sessions_by_sensor_id,
    all_live_response_sessions,
    get_session_commands,
    get_command_result,
    get_file_content,
)
from cbinterface.commands import (
    PutFile,
    ProcessListing,
    GetFile,
    ListRegKeyValues,
    RegKeyValue,
    ExecuteCommand,
    ListDirectory,
    WalkDirectory,
    LogicalDrives,
    DeleteFile,
    KillProcessByID,
    KillProcessByName,
    DeleteRegistryKeyValue,
    DeleteRegistryKey,
    SetRegKeyValue,
    CreateRegKey,
    GetSystemMemoryDump,
)
from cbinterface.response.enumerations import logon_history
from cbinterface.config import get_playbook_map
from cbinterface.scripted_live_response import build_playbook_commands, build_remediation_commands

LOGGER = logging.getLogger("cbinterface.response.cli")


def add_response_arguments_to_parser(subparsers: argparse.ArgumentParser) -> None:
    """Given an argument parser subparser, build a response specific parser."""
    # sensor query parser
    parser_sensor = subparsers.add_parser(
        "sensor-query",
        aliases=["sq"],
        help="Execute a sensor query (Response). Valid search fields: 'ip', 'hostname', and 'groupid'",
    )
    parser_sensor.add_argument("sensor_query", help="the sensor query you'd like to execute")
    parser_sensor.add_argument(
        "-nw",
        "--no-warnings",
        action="store_true",
        default=False,
        help="Don't warn before printing large query results",
    )
    parser_sensor.add_argument(
        "-ad",
        "--all-details",
        action="store_true",
        default=False,
        help="Print all available process info (all fields).",
    )

    # response watchlist parser
    parser_watchlist = subparsers.add_parser(
        "response_watchlist", aliases=["rwl"], help="Work with response watchlists."
    )
    parser_watchlist.add_argument("-l", "--list-watchlists", action="store_true", help="Print all watchlists.")
    parser_watchlist.add_argument(
        "-q", "--query-watchlists", action="store", help="filter watchlists by watchlist query"
    )
    parser_watchlist.add_argument(
        "-json",
        "--watchlists-to-json",
        action="store_true",
        help="Convert watchlists to json and print to stdout.",
    )
    parser_watchlist.add_argument(
        "--watchlist-names-from-stdin", action="store_true", help="read a list of watchlist names from stdin to load."
    )


def execute_response_arguments(cb: CbResponseAPI, args: argparse.Namespace) -> bool:
    """The logic to execute response specific command line arguments.

    Args:
        cb: CbResponseAPI
        args: parsed argparse namespace
    Returns:
        True or None on success, False on failure.
    """

    if not isinstance(cb, CbResponseAPI):
        LOGGER.critical(f"expected CbResponseAPI but got '{type(cb)}'")
        return False

    # Sensor Quering #
    if args.command and (args.command == "sensor-query" or args.command == "sq"):
        LOGGER.info(f"searching {args.environment} environment for sensor query: {args.sensor_query}...")

        sensors = make_sensor_query(cb, args.sensor_query)
        if not sensors:
            return None

        # don't display large results by default
        print_results = True
        if not args.no_warnings and len(sensors) > 10:
            prompt = "Print all results? (y/n) [y] "
            print_results = input_with_timeout(prompt, default="y")
            print_results = True if print_results.lower() == "y" else False

        if len(sensors) > 0 and print_results:
            print("\n------------------------- SENSOR RESULTS -------------------------")
            for sensor in sensors:
                if args.all_details:
                    print()
                    print(sensor)
                else:
                    print(sensor_info(sensor))
            print()
        return True

    # Watchlists #
    if args.command and (args.command == "response_watchlist" or args.command == "rwl"):
        watchlists = watchlist_names = []
        if args.query_watchlists:
            watchlists = query_watchlists(cb, args.query_watchlists)
        elif args.list_watchlists:
            watchlists = get_all_watchlists(cb)

        if args.watchlist_names_from_stdin:
            watchlist_names = [line.strip() for line in sys.stdin]

        if args.watchlists_to_json:
            if watchlists:
                print(json.dumps(these_watchlists_to_list_dict(cb, [wl.name for wl in watchlists])))
            if watchlist_names:
                print(json.dumps(these_watchlists_to_list_dict(cb, watchlist_names)))
            return
        elif len(watchlists) > 0:
            print("\n------------------------- WATCHLISTS -------------------------")
            for wl in watchlists:
                print(wl)

    # Process Quering #
    if args.command and (args.command.startswith("q") or args.command == "pq"):
        LOGGER.info(f"searching {args.environment} environment..")
        args.start_time = (
            datetime.datetime.strptime(args.start_time, "%Y-%m-%d %H:%M:%S") if args.start_time else args.start_time
        )
        args.last_time = (
            datetime.datetime.strptime(args.last_time, "%Y-%m-%d %H:%M:%S") if args.last_time else args.last_time
        )
        processes = make_process_query(
            cb, args.query, start_time=args.start_time, last_time=args.last_time, raise_exceptions=False
        )

        if args.facets:
            LOGGER.info("getting facet data...")
            print_facet_histogram(processes.facets())

        # don't display large results by default
        print_results = True
        if not args.no_warnings and len(processes) > 10:
            prompt = "Print all results? (y/n) [y] "
            print_results = input_with_timeout(prompt, default="y")
            print_results = True if print_results.lower() == "y" else False

        if len(processes) > 0 and print_results:
            print("\n------------------------- QUERY RESULTS -------------------------")
            for proc in processes:
                print("  -------------------------")
                if args.all_details:
                    print(proc)
                else:
                    print_process_info(proc, raw_print=args.all_details, header=False)

        return True

    # Enumerations #
    if args.command and args.command == "enumerate":
        if args.logon_history:
            logon_history(cb, args.logon_history)
            return

    # Process Inspection #
    if args.command and (args.command.lower() == "inspect" or args.command.lower().startswith("proc")):
        process_id = args.process_guid_options
        process_segment = None
        if "/" in args.process_guid_options:
            if not args.process_guid_options.count("/") == 1:
                LOGGER.error(f"process guid/segement format error: {args.process_guid_options}")
                return False
            process_id, process_segment = args.process_guid_options.split("/")
            if not re.match("[0-9]{13}", process_segment):
                LOGGER.error(f"{process_segment} is not in the form of a process segment.")
                return False
            process_segment = int(process_segment)
        if not is_uuid(process_id):
            LOGGER.error(f"{process_id} is not in the form of a globally unique process id (GUID/UUID).")
            return False

        try:
            proc = Process(cb, process_id, force_init=True)
            if process_segment and process_segment not in proc.get_segments():
                LOGGER.warning(f"segment '{process_segment}' does not exist. Setting to first segment.")
                process_segment = None
            proc.current_segment = process_segment
        except ObjectNotFoundError:
            LOGGER.warning(f"ObjectNotFoundError - process data does not exist.")
            return False
        except Exception as e:
            LOGGER.error(f"problem finding process: {e}")
            return False

        all_inspection_args = [iarg for iarg in vars(args).keys() if iarg.startswith("inspect_")]
        set_inspection_args = [
            iarg for iarg, value in vars(args).items() if iarg.startswith("inspect_") and value is True
        ]
        if not set_inspection_args:
            LOGGER.debug(f"seting all inspection arguments.")
            for iarg in all_inspection_args:
                args.__setattr__(iarg, True)

        if args.json:
            print(json.dumps(process_to_dict(proc, max_segments=args.segment_limit), default=str))
            return

        if args.walk_and_inspect_tree:
            inspect_process_tree(
                proc,
                info=args.inspect_proc_info,
                filemods=args.inspect_filemods,
                netconns=args.inspect_netconns,
                regmods=args.inspect_regmods,
                modloads=args.inspect_modloads,
                crossprocs=args.inspect_crossprocs,
                children=args.inspect_children,
                raw_print=args.raw_print_events,
            )
            return True
        # else
        if args.inspect_process_ancestry:
            print_ancestry(proc)
        if args.inspect_process_tree:
            print_process_tree(proc)
        if args.inspect_proc_info:
            print_process_info(proc, raw_print=args.raw_print_events)
        if args.inspect_filemods:
            print_filemods(proc, current_segment_only=bool(process_segment), raw_print=args.raw_print_events)
        if args.inspect_netconns:
            print_netconns(proc, current_segment_only=bool(process_segment), raw_print=args.raw_print_events)
        if args.inspect_regmods:
            print_regmods(proc, current_segment_only=bool(process_segment), raw_print=args.raw_print_events)
        if args.inspect_modloads:
            print_modloads(proc, current_segment_only=bool(process_segment), raw_print=args.raw_print_events)
        if args.inspect_crossprocs:
            print_crossprocs(proc, current_segment_only=bool(process_segment), raw_print=args.raw_print_events)
        if args.inspect_children:
            print_childprocs(proc, current_segment_only=bool(process_segment), raw_print=args.raw_print_events)

    # Live Response Actions #
    if args.command and (args.command.lower() == "lr" or args.command.lower().startswith("live")):
        # create a LR session manager
        session_manager = CustomLiveResponseSessionManager(cb, custom_session_keepalive=True)
        # store a list of commands to execute on this sensor
        commands = []

        try:
            sensor = Sensor(cb, args.name_or_id, force_init=True)
        except ObjectNotFoundError:
            LOGGER.info(f"searching for sensor...")
            sensor = find_sensor_by_hostname(cb, args.name_or_id)

        if not sensor:
            LOGGER.info(f"could not find a sensor.")
            return None

        if args.execute_command:
            # XXX expand this for more flexibiliy by making an execute parser
            # that can accept more arugments to pass to ExecuteCommand
            cmd = ExecuteCommand(args.execute_command)
            commands.append(cmd)
            LOGGER.info(f"recorded command: {cmd}")

        if args.sensor_isolation_toggle:
            result = None
            state = "isolated" if sensor.is_isolating else "unisolated"
            desired_state = "unisolated" if sensor.is_isolating else "isolated"
            LOGGER.info(
                f"sensor {sensor.id}:{sensor.hostname} is currently {state}. Changing state to: {desired_state}"
            )
            if sensor.is_isolating:
                result = sensor.unisolate()
            else:
                result = sensor.isolate()
            if result:
                state = "isolated" if sensor.is_isolating else "unisolated"
                LOGGER.info(f"successfully {state} sensor {sensor.id}:{sensor.hostname}")
            else:
                state = "unisolate" if sensor.is_isolating else "isolate"
                LOGGER.error(f"failed to {state} sensor {sensor.id}:{sensor.hostname}")

        # Put File #
        if args.live_response_command and args.live_response_command.lower() == "put":
            cmd = PutFile(args.local_filepath, args.sensor_write_filepath)
            commands.append(cmd)
            LOGGER.info(f"recorded command: {cmd}")

        if args.create_regkey:
            cmd = CreateRegKey(args.create_regkey)
            commands.append(cmd)
            LOGGER.info(f"recorded command: {cmd}")
            if args.set_regkey_value:
                cmd = SetRegKeyValue(args.create_regkey, args.set_regkey_value)
                commands.append(cmd)
                LOGGER.info(f"recorded command: {cmd}")

        # Sensor Collection #
        if args.live_response_command and args.live_response_command.lower() == "collect":
            if args.sensor_info:
                print(sensor_info(sensor))

            if args.process_list:
                cmd = ProcessListing()
                commands.append(cmd)
                LOGGER.info(f"recorded command: {cmd}")

            if args.list_directory:
                cmd = ListDirectory(args.list_directory)
                commands.append(cmd)
                LOGGER.info(f"recorded command: {cmd}")

            if args.walk_directory:
                cmd = WalkDirectory(args.walk_directory)
                commands.append(cmd)
                LOGGER.info(f"recorded command: {cmd}")

            if args.file:
                cmd = GetFile(args.file)
                commands.append(cmd)
                LOGGER.info(f"recorded command: {cmd}")

            if args.regkeypath:
                cmd = ListRegKeyValues(args.regkeypath)
                commands.append(cmd)
                LOGGER.info(f"recorded command: {cmd}")

            if args.regkeyvalue:
                cmd = RegKeyValue(args.regkeyvalue)
                commands.append(cmd)
                LOGGER.info(f"recorded command: {cmd}")

            if args.drives:
                cmd = LogicalDrives()
                commands.append(cmd)
                LOGGER.info(f"recorded command: {cmd}")

            if args.memdump:
                cmd = GetSystemMemoryDump()
                commands.append(cmd)
                LOGGER.info(f"recorded command: {cmd}")

        # Sensor Remediation #
        if args.live_response_command and args.live_response_command == "remediate":
            if args.delete_file_path:
                cmd = DeleteFile(args.delete_file_path)
                commands.append(cmd)
                LOGGER.info(f"recorded command: {cmd}")

            if args.kill_process_name:
                cmd = KillProcessByName(args.kill_process_name)
                commands.append(cmd)
                LOGGER.info(f"recorded command: {cmd}")

            if args.delete_regkeyvalue:
                cmd = DeleteRegistryKeyValue(args.delete_regkeyvalue)
                commands.append(cmd)
                LOGGER.info(f"recorded command: {cmd}")

            if args.delete_entire_regkey:
                cmd = DeleteRegistryKey(args.delete_entire_regkey)
                commands.append(cmd)
                LOGGER.info(f"recorded command: {cmd}")

            if args.remediation_script:
                remediation_commands = build_remediation_commands(args.remediation_script)
                LOGGER.info(f"created {len(remediation_commands)} remediation commands from {args.remediation_script}")
                commands.extend(remediation_commands)

        # Playbook execution #
        if args.live_response_command and (
            args.live_response_command.startswith("play") or args.live_response_command == "pb"
        ):
            if args.playbook_configpath:
                playbook_commands = build_playbook_commands(args.playbook_configpath)
                commands.extend(playbook_commands)
                LOGGER.info(f"loaded {len(playbook_commands)} playbook commands.")
            if args.playbook_name:
                playbook_data = get_playbook_map()[args.playbook_name]
                playbook_path = playbook_data["path"]
                playbook_commands = build_playbook_commands(playbook_path)
                commands.extend(playbook_commands)
                LOGGER.info(f"loaded {len(playbook_commands)} playbook commands.")

        # Handle LR commands #
        if commands:
            timeout = 1200  # default 20 minutes (same used by Cb)
            if not is_sensor_online(sensor):
                # Decision point: if the sensor is NOT online, give the analyst and option to wait
                LOGGER.warning(f"{sensor.id}:{sensor.hostname} is offline.")
                prompt = "Would you like to wait for the host to come online? (y/n) [y] "
                wait = input_with_timeout(prompt, default="y")
                wait = True if wait.lower() == "y" else False
                if not wait:
                    return None
                prompt = "How many days do you want to wait? [Default is 7 days] "
                timeout = input_with_timeout(prompt, default=7)
                if isinstance(timeout, str):
                    timeout = int(timeout)
                if timeout > 30:
                    LOGGER.warning(f"{timeout} days is a long time. Restricting to max of 30 days.")
                    timeout = 30

                # 86400 seconds in a day
                timeout = timeout * 86400

            if not session_manager.wait_for_active_session(sensor, timeout=timeout):
                LOGGER.error(f"reached timeout waiting for active session.")
                return False

            # we have an active session, issue the commands.
            for command in commands:
                session_manager.submit_command(command, sensor)

        if session_manager.commands:
            # Wait for issued commands to complete and process any results.
            session_manager.process_completed_commands()

    # Direct Session Interaction #
    if args.command and args.command.startswith("sess"):
        if args.list_sensor_sessions:
            print(
                json.dumps(
                    sensor_live_response_sessions_by_sensor_id(cb, args.list_sensor_sessions), indent=2, sort_keys=True
                )
            )

        if args.get_session_command_list:
            print(json.dumps(get_session_commands(cb, args.get_session_command_list), indent=2, sort_keys=True))

        if args.list_all_sessions:
            print(json.dumps(all_live_response_sessions(cb), indent=2, sort_keys=True))

        if args.get_session:
            print(json.dumps(get_session_by_id(cb, args.get_session), indent=2, sort_keys=True))

        if args.close_session:
            session_manager = CustomLiveResponseSessionManager(cb)
            session_manager._close_session(args.close_session)
            print(json.dumps(get_session_by_id(cb, args.close_session), indent=2, sort_keys=True))

        if args.get_command_result:
            session_id, command_id = args.get_command_result.split(":", 1)
            print(json.dumps(get_command_result(cb, session_id, command_id), indent=2, sort_keys=True))

        if args.get_file_content:
            session_id, file_id = args.get_file_content.split(":", 1)
            get_file_content(cb, session_id, file_id)

    return True



"""All things LR sessions commands and command management."""

import os
import json
import time
import logging
from datetime import timedelta
from typing import Union
from concurrent.futures import Future

from cbapi.live_response_api import CbLRSessionBase

LOGGER = logging.getLogger("cbinterface.command")

# the cbinterface directory
BASE_DIR = os.path.dirname(os.path.abspath(__file__))


class BaseSessionCommand:
    """For storing and managing session commands.

    Session Commands are 'jobs' with concurrent.futures.
    """

    def __init__(self, description, placeholders={}, post_completion_command=None):
        self.description = description
        self.future = None
        self.session_id = None
        # store a copy of session data
        self.session_data = {}
        self._sensor_id = None
        self._exception = None
        self._result = None
        self._hostname = None
        self.placeholders = placeholders
        self.post_completion_command = post_completion_command

    def fill_placeholders(self, string_item: str, placeholders={}):
        # fill common placeholders
        placeholders = placeholders if placeholders else self.placeholders
        placeholders["HOSTNAME"] = placeholders.get("HOSTNAME", self.hostname)
        placeholders["SENSOR_ID"] = placeholders.get("SENSOR_ID", self.sensor_id)
        placeholders["DEVICE_ID"] = placeholders.get("DEVICE_ID", self.sensor_id)
        placeholders["WORK_DIR"] = placeholders.get("WORK_DIR", "C:\\Program Files")
        string_item = string_item.format(**placeholders)
        return string_item

    @property
    def hostname(self):
        # NOTE: Appears this is set to None for psc. So, custom
        # session manager sets it right before job submission.
        return self.session_data.get("hostname") or self._hostname

    @property
    def sensor_id(self):
        return self.session_data.get("sensor_id") or self._sensor_id

    @property
    def initiatied(self):
        """True if future exists."""
        if isinstance(self.future, Future):
            return True
        return False

    @property
    def done(self):
        if self.future and self.future.done():
            return True
        return False

    @property
    def exception(self):
        if self.done:
            self._exception = self.future.exception()
        return self._exception

    @property
    def has_result(self):
        if self.done and self.exception is None:
            return True
        return False

    @property
    def result(self):
        if self.has_result:
            return self.future.result()
        return None

    @property
    def status(self):
        if not self.initiatied:
            return "not submitted"
        if self.done:
            if self.exception:
                return "error"
            if self.has_result:
                return "complete"
        return "pending"

    def run(self):
        """The function to implement the live response command logic."""
        raise NotImplemented()

    def process_result(self):
        """Implement logic to process any results."""
        pass

    def execute_post_completion(self):
        if not self.post_completion_command:
            return None
        if self.status != "complete":
            return False
        self.post_completion_command = self.fill_placeholders(self.post_completion_command)
        if self.post_completion_command.startswith("tools/"):
            self.post_completion_command = f"{BASE_DIR}/{self.post_completion_command}"
        LOGGER.info(f"executing post completion command: {self.post_completion_command}")
        import shlex, subprocess

        try:
            args = shlex.split(self.post_completion_command)
            return subprocess.run(args=args)
        except Exception as e:
            LOGGER.error(f"caught exception executing post completion command: {e}")
            return False

    def __str__(self):
        txt = f"LrSessionCommand: {self.description}"
        if self.hostname:
            txt += f" - device:{self.hostname}"
        if self.session_id:
            txt += f" - session:{self.session_id}"
        txt += f" - status:{self.status}"
        return txt


###############################
# General or Generic Commands #
###############################
class PutFile(BaseSessionCommand):
    """Get process listing via Live Response."""

    def __init__(self, local_filepath, sensor_write_filepath, **kwargs):
        super().__init__(description="Put file on device", **kwargs)
        self.local_filepath = local_filepath
        self.sensor_write_filepath = sensor_write_filepath

    def run(self, session: CbLRSessionBase):
        self.sensor_write_filepath = self.fill_placeholders(self.sensor_write_filepath)
        try:
            with open(self.local_filepath, "rb") as fp:
                data = fp.read()
            return session.put_file(data, self.sensor_write_filepath)
        except Exception as e:
            LOGGER.error(f"couldn't put file: {e}")
            return False

    def process_result(self):
        if not self.result:
            return False
        # it worked if execution makes it here
        LOGGER.info(f"put '{self.sensor_write_filepath}' on {self.hostname} via session {self.session_id}")
        if self.post_completion_command:
            self.execute_post_completion()
        return True


class LogicalDrives(BaseSessionCommand):
    """List logical drives on this sensor."""

    def __init__(self):
        super().__init__(description="List logical drives")

    def run(self, session: CbLRSessionBase):
        return session.session_data.get("drives", [])

    def process_result(self):
        print(f"\n + Logical Drives on {self.hostname}: ")
        for drive in self.result:
            print(f"    {drive}")


class CreateRegKey(BaseSessionCommand):
    """Create a registry key on the remote machine."""

    def __init__(self, regkey):
        super().__init__(description=f"Create A Registry Key @ '{regkey}'")
        self.regkey = regkey

    def run(self, session: CbLRSessionBase):
        return session.create_registry_key(self.regkey)

    def process_result(self):
        LOGGER.info(f"Created'{self.regkey}' on {self.hostname}")


class SetRegKeyValue(BaseSessionCommand):
    """Set a registry key value on the remote machine."""

    def __init__(self, regkey, value):
        super().__init__(description=f"Set RegKey '{regkey}' value '{value}'")
        self.regkey = regkey
        self.value = value

    def run(self, session: CbLRSessionBase):
        return session.set_registry_value(self.regkey, self.value)

    def process_result(self):
        LOGGER.info(f"Set '{self.regkey}' value='{self.value}' on {self.hostname}")


class ExecuteCommand(BaseSessionCommand):
    """Create a new process on the remote machine with the specified command string.

    Args:
        command (str): Command string used for the create process operation.
        wait_for_output (bool): True to block on output from the new process (execute in foreground).
            This will also set wait_for_completion (below).
        remote_output_file_name (str): The remote output file name used for process output.
        working_directory (str): The working directory of the create process operation.
        wait_timeout (int): Timeout used for this command.
        wait_for_completion (bool): True to wait until the process is completed before returning.
        print_results (bool): Print results to console if True.
        write_results_path: Path to write the results. Default is to write nothing.

    Returns:
        str: The output of the process.
    """

    def __init__(
        self,
        command: str,
        wait_for_output=True,
        remote_output_file_name=None,
        working_directory=None,
        wait_timeout=60,
        wait_for_completion=True,
        print_results=True,
        write_results_path=False,
        **kwargs,
    ):
        super().__init__(description=f"Execute {command}", **kwargs)
        self._command_string = command
        self.wait_for_output = wait_for_output
        self.remote_output_file_name = remote_output_file_name
        self.working_directory = working_directory
        self.wait_timeout = wait_timeout
        self.wait_for_completion = wait_for_completion
        self.print_results = print_results
        self.write_results_path = write_results_path
        self.start_time = None
        self.elapsed_time = None

    def run(self, session: CbLRSessionBase):
        self._command_string = self.fill_placeholders(self._command_string)
        self.start_time = time.time()
        session.create_process(
            self._command_string,
            wait_for_output=self.wait_for_output,
            remote_output_file_name=self.remote_output_file_name,
            working_directory=self.working_directory,
            wait_timeout=self.wait_timeout,
            wait_for_completion=self.wait_for_completion,
        )
        self.elapsed_time = timedelta(seconds=(time.time() - self.start_time))

    def process_result(self):
        LOGGER.debug(f"{self} took {self.elapsed_time} to return.")
        if self.post_completion_command:
            self.execute_post_completion()
        if not self.result:
            if self.wait_for_output:
                LOGGER.warning("Expected output, but did not receive results.")
                return False
            LOGGER.info(f"successfully executed '{self._command_string}'. no results returned.")
            return True
        if self.print_results:
            print("\n-------------------------")
            print(self.result.decode("utf-8"))
            print("\n-------------------------")
            print()
        if self.write_results_path:
            self.write_results_path = self.fill_placeholders(self.write_results_path)
            if os.path.exists(self.write_results_path):
                LOGGER.info(f"overriting existing file: {self.write_results_path}")
            with open(self.write_results_path, "wb") as fp:
                fp.write(self.result)
        return True


#######################
# Collection Commands #
#######################
class ProcessListing(BaseSessionCommand):
    """Get process listing via Live Response."""

    def __init__(self):
        super().__init__(description="process listing")

    def run(self, session: CbLRSessionBase):
        return session.list_processes()

    def process_result(self):
        # just print the results as json
        print(json.dumps(self.result, indent=2, sort_keys=True))


class ListDirectory(BaseSessionCommand):
    """List the contents of a directory."""

    def __init__(self, dir_path: str):
        super().__init__(description=f"List Directory @ {dir_path}")
        self.dir_path = dir_path

    def run(self, session: CbLRSessionBase):
        return session.list_directory(self.dir_path)

    def process_result(self):
        results = self.result
        if not results:
            LOGGER.info("no results.")
            return
        print(json.dumps(results, indent=2, sort_keys=True))


class WalkDirectory(BaseSessionCommand):
    """Perform a full directory walk with recursion into subdirectories."""

    def __init__(self, dir_path: str, followlinks=False):
        super().__init__(description=f"Walk Directory @ {dir_path}")
        self.dir_path = dir_path
        self.followlinks = followlinks

    def run(self, session: CbLRSessionBase):
        return session.walk(self.dir_path, followlinks=self.followlinks)

    def process_result(self):
        if not self.result:
            LOGGER.warning("no results.")
            return
        print(f"Recursive top-down directory listing for '{self.dir_path}':")
        for name, subdir_names, filenames in self.result:
            print(f" + Directory: {name}")
            for subdir in subdir_names:
                print(f"    SubDirectory: {subdir}")
            for fn in filenames:
                print(f"    File: {fn}")


class ListRegKeyValues(BaseSessionCommand):
    """List all registry values from the specified registry key."""

    def __init__(self, regkeypath: str, return_json=False):
        super().__init__(description=f"List Registry Keys and Values @ {regkeypath}")
        self.regkeypath = regkeypath
        self.return_json = return_json

    def run(self, session: CbLRSessionBase):
        return session.list_registry_keys(self.regkeypath)
        # return session.list_registry_keys_and_values(self.regkeypath)

    def process_result(self):
        results = self.result
        if not results:
            LOGGER.info("no results.")
            return None
        if self.return_json:
            return json.dumps(results, indent=2, sort_keys=True)
        print(f"\nRegistryKey Listing: {self.regkeypath}")
        for result in results:
            print("\n\t-------------------------")
            print(f"\tName: {result['value_name']}")
            print(f"\tType: {result['value_type']}")
            print(f"\tData: {result['value_data']}")
        print()


class RegKeyValue(BaseSessionCommand):
    """Get the associated value of the specified registry key."""

    # TODO: add write_results option and make print_results an option

    def __init__(self, regkeyvalue: str):
        super().__init__(description=f"Get Registry Key Value @ {regkeyvalue}")
        self.regkeyvalue = regkeyvalue

    def run(self, session: CbLRSessionBase):
        return session.get_registry_value(self.regkeyvalue)

    def process_result(self):
        # print(json.dumps(self.result, indent=2, sort_keys=True))
        print("\n\t-------------------------")
        print(f"\tName: {self.result['value_name']}")
        print(f"\tType: {self.result['value_type']}")
        print(f"\tData: {self.result['value_data']}")
        print()


class GetSystemMemoryDump(BaseSessionCommand):
    """Perform a memory dump operation on the sensor.

    NOTE: Not a fan of Cb's implementation.
    """

    def __init__(self, local_filename: str = "", compress=True):
        super().__init__(description=f"Dump System Memory")
        self.local_filename = local_filename
        self.compress = compress
        self._memdump_id = None
        self._cb = None

    def run(self, session: CbLRSessionBase):
        # store a pointer to the CbR object for later
        self._cb = session._cb
        if not self.local_filename:
            self.local_filename = f"{self.sensor_id}_{self.hostname}.cb.memdump"
        if self.compress:
            self.local_filename += ".zip"
        dump_object = session.start_memdump(compress=self.compress)
        self._memdump_id = dump_object.memdump_id
        dump_object.wait()
        dump_object.get(local_filename)
        dump_object.delete()

    def process_result(self):
        from cbinterface.response.sessions import get_command_result

        # should only make it here if an error was not raise
        # check to see if the command has success status with server
        # and if the local file exists
        memdump_cmd = get_command_result(self._cb, self.session_id, self._memdump_id)
        if memdump["status"] != "complete":
            LOGGER.error(f"problem completing memory dump: command status: {memdump['status']}")
            return False
        if os.path.exits(self.local_filename):
            LOGGER.info(f" +  wrote: {self.local_filename}")
            return True
        else:
            LOGGER.error(f"Memory dump completed but failed to get a local copy of the memory dump.")
            return False


class GetFile(BaseSessionCommand):
    """Object that retrieves a file via Live Response."""

    def __init__(self, file_path, output_filename: Union[str, bool] = None, **kwargs):
        """
        Initialize the GetFile command.

        Args:
            file_path (str): The file path to be fetched.
            output_filename: optional path to write the file content.
        Returns:
            True on success, False on failure.
        """
        super().__init__(description=f"getFile @ '{file_path}'", **kwargs)
        self._file_path = file_path

        self.output_filename = output_filename

    def run(self, session: CbLRSessionBase):
        """
        Execute the file transfer.
        Args:
            session (CbLRSessionBase): The Live Response session being used.
        Returns:
            File content
        """
        if "{WILDMATCH}" in self._file_path:
            # split on "{WILDMATCH}" and search for the first match to collect
            from cbinterface.helpers import get_os_independent_filepath

            file_path_parts = [self.fill_placeholders(fpp) for fpp in self._file_path.split("{WILDMATCH}")]
            dir_path = get_os_independent_filepath(file_path_parts[0]).parent
            dir_path = f"{dir_path}\\" if "\\" in str(dir_path) else f"{dir_path}/"

            LOGGER.info(f"attempting to find item at '{dir_path}' like {file_path_parts}")
            for item in session.list_directory(dir_path):
                if item["attributes"] == "DIRECTORY":
                    continue
                if [part for part in file_path_parts if part in item["filename"]]:
                    LOGGER.info(f"found potential match: {item['filename']}")
                    self._file_path = f"{dir_path}{item['filename']}"
                    break

        self._file_path = self.fill_placeholders(self._file_path)
        return session.get_raw_file(self._file_path)

    def process_result(self):
        """Write the results to a local file."""
        from cbinterface.helpers import get_os_independent_filepath

        if self.output_filename is None:
            filepath = get_os_independent_filepath(self._file_path)
            hostname_part = f"{self.hostname}_" if self.hostname else ""
            self.output_filename = f"{self.sensor_id}_{hostname_part}{filepath.name}"
        else:
            self.output_filename = self.fill_placeholders(self.output_filename)

        try:
            if os.path.exists(self.output_filename):
                LOGGER.debug(f"{self.output_filename} already exists. appending epoch time")
                _now = str(time.time())
                _now = _now[: _now.rfind(".")]
                self.output_filename = f"{_now}_{self.output_filename}"
            with open(self.output_filename, "wb") as fp:
                content_handle = self.result
                fp.write(content_handle.read())
                content_handle.close()
            if os.path.exists(self.output_filename):
                LOGGER.info(f"wrote: {self.output_filename}")
            if self.post_completion_command:
                self.execute_post_completion()
            return True
        except Exception as e:
            LOGGER.error(f"problem getting file content: {e}")
            return False


########################
# Remediaiton Commands #
########################
class DeleteFile(BaseSessionCommand):
    """Object that deletes a file via Live Response."""

    def __init__(self, file_path):
        """
        Delete the specified file name on the remote machine.
        Args:
            filename (str): Name of the file to be deleted.
        """
        super().__init__(description=f"Delete File @ '{file_path}'")
        self._file_path = file_path

    def run(self, session: CbLRSessionBase):
        return session.delete_file(self._file_path)

    def process_result(self):
        LOGGER.info(f"deleted '{self._file_path}' on {self.hostname} via session {self.session_id}")
        return True


class KillProcessByID(BaseSessionCommand):
    """
    Terminate a process by process id.
    Args:
        pid (int): Process ID to be terminated.
    Returns:
        bool: True if success, False if failure.
    """

    def __init__(self, pid):
        super().__init__(description=f"Kill Process with ID={pid}")
        self.pid = pid

    def run(self, session: CbLRSessionBase):
        return session.kill_process(self.pid)

    def process_result(self):
        if self.result is True:
            LOGGER.info(f"Killed process id {self.pid} on {self.hostname} via session {self.session_id}")
            return True
        else:
            LOGGER.error(f"failed to kill process id {self.pid} on {self.hostname} via session {self.session_id}")
            return self.result


class DeleteRegistryKeyValue(BaseSessionCommand):
    """Delete a registry value on the remote machine."""

    def __init__(self, regkeyvalue):
        super().__init__(description=f"Delete Registry Key:Value @ '{regkeyvalue}'")
        self.regkeyvalue = regkeyvalue

    def run(self, session: CbLRSessionBase):
        return session.delete_registry_value(self.regkeyvalue)

    def process_result(self):
        LOGGER.info(f"Deleted '{self.regkeyvalue}' on {self.hostname}")


class DeleteRegistryKey(BaseSessionCommand):
    """Delete a registry key and all it's values."""

    def __init__(self, regkey):
        super().__init__(description=f"Delete Registry Key @ '{regkey}'")
        self.regkey = regkey

    def run(self, session: CbLRSessionBase):
        return session.delete_registry_key(self.regkey)

    def process_result(self):
        LOGGER.info(f"Deleted '{self.regkey}' on {self.hostname}")


class KillProcessByName(BaseSessionCommand):
    """
    Terminate a process by process name.
    Args:
        process_name (str): Process name(s) to be terminated.
    Returns:
        bool: True if success, False if failure.
    """

    def __init__(self, process_name):
        super().__init__(description=f"Kill Processes with name like '{process_name}'")
        self.pname = process_name
        self.nested_commands = {}

    def run(self, session: CbLRSessionBase):
        from cbinterface.helpers import get_os_independent_filepath

        for process in session.list_processes():
            filepath = get_os_independent_filepath(process["path"])
            if self.pname.lower() in filepath.name.lower():
                LOGGER.info(f"found process to kill: {process['path']} - pid={process['pid']}")
                self.nested_commands[process["pid"]] = session.kill_process(process["pid"])

        return True

    def process_result(self):
        if self.result:
            for pid, killed in self.nested_commands.items():
                if killed:
                    LOGGER.info(f"successfully killed: pid={pid}")
                else:
                    LOGGER.warning(f"failed to kill: pid={pid}")


""" Not used: left as an example of recursive command."""


class RecursiveKillProcessByName(BaseSessionCommand):
    """Recursivly terminate a process by process name.

    First, list processes and find matching process names. Next,
    Create a local CustomLiveResponseSessionManager and create
    commands (managed via threads) to kill the matching processes by PID.
    """

    def __init__(self, process_name):
        super().__init__(description=f"Kill Processes with name like '{process_name}'")
        self.pname = process_name
        self.local_session_manager = None

    def run(self, session: CbLRSessionBase):
        from cbinterface.helpers import get_os_independent_filepath
        from cbinterface.response.sessions import CustomLiveResponseSessionManager

        self.local_session_manager = CustomLiveResponseSessionManager(session._cb)
        for process in session.list_processes():
            filepath = get_os_independent_filepath(process["path"])
            if self.pname in filepath.name:
                LOGGER.info(f"found process to kill: {process['path']} - pid={process['pid']}")
                cmd = KillProcessByID(process["pid"])
                self.local_session_manager.submit_command(cmd, self.sensor_id)

        return True

    def process_result(self):
        if self.result:
            self.local_session_manager.process_completed_commands()



"""All things LR sessions."""

import io
import os
import time
import logging
import threading

from cbapi.psc import Device
from cbapi.psc.threathunter import CbThreatHunterAPI
from cbapi.psc.cblr import (
    LiveResponseSession,
    LiveResponseSessionManager,
    LiveResponseJobScheduler,
    WorkItem,
    JobWorker,
)
from cbapi.errors import ObjectNotFoundError, TimeoutError

# from cbapi.live_response_api import CbLRManagerBase, WorkItem, poll_status

from typing import List, Union

from cbinterface.commands import BaseSessionCommand
from cbinterface.psc.device import is_device_online

LOGGER = logging.getLogger("cbinterface.psc.session")

CBLR_BASE = "/integrationServices/v3/cblr"


class CustomLiveResponseJobScheduler(LiveResponseJobScheduler):
    def __init__(self, cb, psc_cb, max_workers=10):
        self.psc_cb = psc_cb
        super().__init__(cb, max_workers=10)

    def _spawn_new_workers(self):
        if len(self._job_workers) >= self._max_workers:
            return

        schedule_max = self._max_workers - len(self._job_workers)

        devices = [
            s for s in self.psc_cb.select(Device) if s.id in self._unscheduled_jobs and s.id not in self._job_workers
        ]
        # and is_device_online(s)]

        devices_to_schedule = devices[:schedule_max]
        LOGGER.debug("Spawning new workers to handle these devices: {0}".format(devices_to_schedule))
        for device in devices_to_schedule:
            LOGGER.debug("Spawning new JobWorker for device id {0}".format(device.id))
            self._job_workers[device.id] = JobWorker(self._cb, device.id, self.schedule_queue)
            self._job_workers[device.id].start()


class CustomLiveResponseSessionManager(LiveResponseSessionManager):
    def __init__(self, cb, timeout=30, custom_session_keepalive=False):
        # First, get a CB object with the LR API permissions
        cblr = CbThreatHunterAPI(url=cb.credentials.url, token=cb.credentials.lr_token, org_key=cb.credentials.org_key)
        super().__init__(cblr, timeout=timeout, keepalive_sessions=False)
        # so now self._cb == cblr -- store a reference to the regular cb
        self.psc_cb = cb

        if custom_session_keepalive:
            self._cleanup_thread = threading.Thread(target=self._keep_active_sessions_alive_thread)
            self._cleanup_thread.daemon = True
            self._cleanup_thread.start()

        # for storing initiatied commands
        self.commands = []

    def get_session(self, device: Device):
        """Get or create LR session."""
        session_data = self._cb.post_object(f"{CBLR_BASE}/session/{device.id}", {"sensor_id": device.id}).json()
        session_id = session_data["id"]
        LOGGER.debug(f"got session id={session_id} with status={session_data['status']}")
        self._sessions[device.id] = self.cblr_session_cls(self, session_id, device.id, session_data=session_data)
        return self._sessions[device.id]

    def wait_for_active_session(self, device: Device, timeout=86400):
        """Return active session or None.

        Default timeout is 7 days.
        """
        LOGGER.info(
            f"attempting to get active session on device {device.id} (hostname:{device.name}) for up to {timeout/60} minutes"
        )
        start_time = time.time()
        session = None
        status = None
        while status != "ACTIVE" and time.time() - start_time < timeout:
            if not is_device_online(device):
                LOGGER.debug(f"waiting for device {device.id} to come online...")
                time.sleep(1)
                continue
            session = self.get_session(device)
            status = session.session_data["status"]
            time.sleep(0.5)

        if session and is_session_active(session):
            LOGGER.info(f"got active session {session.session_id}.")
        return session

    def submit_command(self, command: BaseSessionCommand, device: Union[int, Device]):
        """
        Create a new job to be executed as a Live Response.

        Args:
            command (BaseSessionCommand): The job to be scheduled.
            device (Device): Device to execute job on.
        Returns:
            Future: A reference to the running job.
        """
        assert isinstance(command, BaseSessionCommand)

        device_id = device
        if isinstance(device, Device):
            device_id = device.id
            command._hostname = device.name
        LOGGER.debug(f"submitting {command} to {device_id}")

        if self._job_scheduler is None:
            # spawn the scheduler thread
            self._job_scheduler = CustomLiveResponseJobScheduler(self._cb, self.psc_cb)
            self._job_scheduler.start()

        if device_id not in self._sessions:
            device = Device(self._cb, device_id, force_init=True)
            active_session = self.active_session(device)
            if active_session is None:
                self.wait_for_active_session(device)

        work_item = WorkItem(command.run, device_id)
        self._job_scheduler.submit_job(work_item)
        command.future = work_item.future
        command.device_id = device_id
        command.session_id = self._sessions[device_id].session_id
        command.session_data = self._sessions[device_id].session_data
        self.commands.append(command)
        return command

    def yield_completed_commands(self):
        """Wait for commands to complete, process results.

        Monitor commands and sessions.
        """
        LOGGER.info(f"waiting for {len(self.commands)} commands to complete ...")
        while self.commands:
            for cmd in self.commands:
                if not cmd.initiatied:
                    LOGGER.error(f"skipping uninitialized command: {cmd}")
                    self.commands.remove(cmd)
                    continue
                if cmd.exception:
                    LOGGER.error(f"exception for {cmd}: {cmd.exception}")
                    self.commands.remove(cmd)
                    continue
                if not get_session_by_id(self._cb, cmd.session_id):
                    LOGGER.error(f"session {cmd.session_id} is gone. command has gone to the void: {cmd}")
                    self.commands.remove(cmd)
                    continue

                if cmd.has_result:
                    LOGGER.debug(f"yielding {cmd}")
                    yield cmd
                    self.commands.remove(cmd)

                # yield time for completion
                time.sleep(0.7)

    def process_completed_commands(self):
        for cmd in self.yield_completed_commands():
            LOGGER.debug(f"processing => {cmd}")
            cmd.process_result()

    def _keep_active_sessions_alive_thread(self):
        """Used by a thread to ping active sessions so they don't
        close on long running session commands.
        """
        LOGGER.debug("Starting custom Live Response session keepalive and cleanup task")
        while True:
            time.sleep(self._timeout)

            delete_list = []
            with self._session_lock:
                for session in self._sessions.values():
                    if session._refcount == 0:
                        delete_list.append(session.device_id)
                    else:
                        try:
                            if is_session_active(session):
                                LOGGER.info(f"sending keepalive for session {session.session_id}")
                                self._send_keepalive(session.session_id)
                        except ObjectNotFoundError:
                            LOGGER.debug(
                                f"Session {session.session_id} for device {session.device_id} not valid any longer, removing from cache"
                            )
                            delete_list.append(session.device_id)
                        except Exception as e:
                            LOGGER.warning(
                                f"Keepalive on session {session.session_id} (device {session.device_id}) failed with unknown error: {e}"
                            )
                            delete_list.append(session.device_id)

                for device_id in delete_list:
                    self._close_session(self._sessions[device_id].session_id)
                    del self._sessions[device_id]


def all_live_response_sessions(cb: CbThreatHunterAPI) -> List:
    """List all LR sessions still in server memory."""
    return [sesh for sesh in cb.get_object(f"{CBLR_BASE}/session")]


def active_live_response_sessions(cb: CbThreatHunterAPI) -> List:
    """Return active LR sessions."""
    return [sesh for sesh in cb.get_object(f"{CBLR_BASE}/session?active_only=true")]


def device_live_response_sessions(device: Device, active_or_pending=False):
    """Get sessions associated to this device."""
    sessions = [session for session in all_live_response_sessions(device._cb) if session["device_id"] == device.id]
    if active_or_pending:
        return [session for session in sessions if session["status"] == "active" or session["status"] == "pending"]
    return sessions


def device_live_response_sessions_by_device_id(cb: CbThreatHunterAPI, device_id: Union[int, str]):
    """Get sessions associated to this device by device id."""
    if isinstance(device_id, str):
        device_id = int(device_id)
    sessions = [session for session in all_live_response_sessions(cb) if session["device_id"] == device_id]
    return sessions


def get_session_by_id(cb: CbThreatHunterAPI, session_id):
    """Get a LR session object by id."""
    try:
        return cb.get_object(f"{CBLR_BASE}/session/{session_id}")
    except ObjectNotFoundError:
        LOGGER.warning(f"no live resonse session by ID={session_id}")
        return None


def close_session_by_id(cb: CbThreatHunterAPI, session_id):
    """Close a session by ID."""
    return cb.put_object(f"{CBLR_BASE}/session", {"session_id": session_id, "status": "CLOSE"}).json()


def get_session_status(cb: CbThreatHunterAPI, session_id):
    """Return any session status or None."""
    session = get_session_by_id(cb, session_id)
    if session is None:
        return None
    return session["status"]


def is_session_active(session: LiveResponseSession):
    """Return True if session is active."""
    if session.session_data["status"] == "ACTIVE":
        return True
    return False


def get_session_commands(cb: CbThreatHunterAPI, session_id: str):
    """List commands for this session."""
    try:
        return cb.get_object(f"{CBLR_BASE}/session/{session_id}/command")
    except ObjectNotFoundError:
        LOGGER.warning(f"no live resonse session by ID={session_id}")
        return None


def get_command_result(cb: CbThreatHunterAPI, session_id: str, command_id: str):
    """Get results of a LR session command."""
    try:
        return cb.get_object(f"{CBLR_BASE}/session/{session_id}/command/{command_id}")
    except ObjectNotFoundError:
        LOGGER.warning(f"no live resonse session and/or command combination for {session_id}:{command_id}")
        return None


def get_file_content(cb: CbThreatHunterAPI, session_id: str, file_id: str):
    """Get file content stored in LR session and write the file locally."""
    from cbinterface.helpers import get_os_independent_filepath

    try:
        real_session_id, device_id = session_id.split(":", 1)
        filename = f"{real_session_id}_on_{device_id}"
        file_metadata = cb.get_object(f"{CBLR_BASE}/session/{session_id}/file/{file_id}")
        if file_metadata:
            filepath = get_os_independent_filepath(file_metadata["file_name"])
            filename = f"{filename}_{filepath.name}"
        result = cb.session.get(f"{CBLR_BASE}/session/{session_id}/file/{file_id}/content", stream=True)
        if result.status_code != 200:
            LOGGER.error(
                f"got {result.status_code} from server getting file {file_id} content for session {session_id}"
            )
            return
        with open(filename, "wb") as fp:
            for chunk in result.iter_content(io.DEFAULT_BUFFER_SIZE):
                fp.write(chunk)
        if os.path.exists(filename):
            LOGGER.info(f"wrote: {filename}")
        return os.path.exists(filename)
    except ObjectNotFoundError:
        LOGGER.warning(f"no file {file_id} content with session {session_id}")
        return



#!/usr/bin/env python3

import sqlite3
import pandas as pd
import argparse
import sys
import os

parser = argparse.ArgumentParser()
parser.add_argument("history_file", help="Path to Chromium based sqlite history file.")
args = parser.parse_args()

if os.path.exists(args.history_file):
    None
else:
    print("[ERROR] file does not exist.")
    sys.exit(1)

db = sqlite3.connect(args.history_file)
c = db.cursor()

tables = pd.read_sql_query("SELECT name FROM sqlite_master WHERE type='table'", db)
tables = list(tables["name"])

browser = "chromium"
if "chrome" in args.history_file.lower():
    browser = "chrome"
elif "edge" in args.history_file.lower():
    browser = "edge"

host = "_unknown_host"
if "_" in args.history_file:
    host = args.history_file[: args.history_file.find("_")]
part_dir = f"{host}_{browser}_raw_csv_tables"
if not os.path.exists(part_dir):
    os.mkdir(part_dir)

for table in tables:
    try:
        data = pd.read_sql_query(f"SELECT * FROM {table}", db)
        data.to_csv(f"{part_dir}/{browser}_history_{table}.csv")
        if os.path.exists(f"{part_dir}/{browser}_history_{table}.csv"):
            print(f" + WROTE: {part_dir}/{browser}_history_{table}.csv")
    except Exception as e:
        print(f"[WARN] problem selecting from table={table}: {e}")

# Try to make it more human readable
browsing = pd.read_sql_query(
    "SELECT datetime(((visits.visit_time/1000000)-11644473600), 'unixepoch') AS date, urls.url, urls.title FROM urls, visits WHERE urls.id = visits.url;",
    db,
)

browsing.set_index("date", inplace=True)
browsing.to_csv(f"{host}_{browser}_history.csv")
if os.path.exists(f"{host}_{browser}_history.csv"):
    print(f" + WROTE: {host}_{browser}_history.csv")

downloads = None
try:
    downloads = pd.read_sql_query(
        "SELECT datetime(downloads.start_time, 'unixepoch') AS date, downloads.url, downloads.full_path, downloads.received_bytes, downloads.total_bytes FROM downloads;",
        db,
    )
except:
    downloads = pd.read_sql_query(
        "SELECT datetime(((downloads.start_time/1000000)-11644473600), 'unixepoch') AS date, downloads.target_path, downloads_url_chains.url, downloads.received_bytes, downloads.total_bytes \
        FROM downloads, downloads_url_chains WHERE downloads.id = downloads_url_chains.id;",
        db,
    )

downloads.set_index("date", inplace=True)
downloads.to_csv(f"{host}_{browser}_downloads.csv")

if os.path.exists(f"{host}_{browser}_downloads.csv"):
    print(f" + WROTE: {host}_{browser}_downloads.csv")



import pytest
import json
import os

from io import StringIO
from contextlib import redirect_stdout

from cbapi.response import CbResponseAPI, Process
from cbapi.errors import ObjectNotFoundError

from cbinterface.config import set_timezone

HOME_PATH = os.path.dirname(os.path.abspath(__file__))


@pytest.fixture(autouse=True)
def no_requests(monkeypatch):
    """Remove requests.sessions.Session.request for all tests."""
    monkeypatch.delattr("requests.sessions.Session.request")


def fake_cb_response_api(monkeypatch):
    """Fake CbResponseAPI object."""

    def fake_info(self):
        server_info = {}
        with open(f"{HOME_PATH}/test_data/cb_response_server_info.json", "r") as fp:
            server_info = json.load(fp)
        server_info["version"] = "5.0.0"
        return server_info

    monkeypatch.setattr(CbResponseAPI, "info", fake_info)
    return CbResponseAPI(url="https://fakehost", token="N/A", ssl_verify=False)


def get_process(monkeypatch):
    from cbapi.response.models import CbFileModEvent
    from cbapi.response.models import CbNetConnEvent
    from cbapi.response.models import CbRegModEvent
    from cbapi.response.models import CbModLoadEvent
    from cbapi.response.models import CbCrossProcEvent
    from cbapi.response.models import CbChildProcEvent

    def _get_segments():
        proc._segments = initial_data["captured_segments"].keys()
        return proc._segments

    def _refresh():
        return True

    def _require_events():
        proc._full_init = True
        proc._events_loaded = True
        return

    def _retrieve_cb_info():
        return cb.server_info

    def _all_filemods():
        for segment_id in proc.get_segments():
            for fm in proc._events[segment_id]["filemods"]:
                timestamp = fm["timestamp"]
                seq = fm["sequence"]
                yield CbFileModEvent(proc, timestamp, seq, fm)

    def _all_netconns():
        for segment_id in proc.get_segments():
            for nc in proc._events[segment_id]["netconns"]:
                timestamp = nc["timestamp"]
                seq = nc["sequence"]
                yield CbNetConnEvent(proc, timestamp, seq, nc)

    def _all_regmods():
        for segment_id in proc.get_segments():
            for rm in proc._events[segment_id]["regmods"]:
                timestamp = rm["timestamp"]
                seq = rm["sequence"]
                yield CbRegModEvent(proc, timestamp, seq, rm)

    def _all_modloads():
        def _is_signed(self):
            return self["is_signed"]

        modloads = []
        for segment_id in proc.get_segments():
            for ml in proc._events[segment_id]["modloads"]:
                timestamp = ml["timestamp"]
                seq = ml["sequence"]
                monkeypatch.setattr(CbModLoadEvent, "is_signed", _is_signed)
                yield CbModLoadEvent(proc, timestamp, seq, ml)

    def _all_crossprocs():
        for segment_id in proc.get_segments():
            for cp in proc._events[segment_id]["crossprocs"]:
                timestamp = cp["timestamp"]
                seq = cp["sequence"]
                yield CbCrossProcEvent(proc, timestamp, seq, cp)

    def _all_childprocs():
        for segment_id in proc.get_segments():
            for cp in proc._events[segment_id]["children"]:
                timestamp = cp["timestamp"]
                seq = cp["sequence"]
                is_suppressed = cp["is_suppressed"]
                proc_data = cp["proc_data"]
                yield CbChildProcEvent(proc, timestamp, seq, cp, is_suppressed=is_suppressed, proc_data=proc_data)

    def _walk_children(callback, max_depth=1, depth=0):
        if max_depth and depth > max_depth:
            return
        if not proc.terminated:
            try:
                callback(proc, depth=depth)
            except ObjectNotFoundError:
                pass
            else:
                proc.walk_children(callback, max_depth=max_depth, depth=depth + 1)

    def _walk_parents(callback, max_depth=1, depth=0):
        if max_depth and depth > max_depth:
            return
        try:
            parent_proc = proc
            if parent_proc and parent_proc.get("process_pid", -1) != -1:
                callback(parent_proc, depth=depth)
            else:
                return
        except ObjectNotFoundError:
            return
        else:
            parent_proc.walk_parents(callback, max_depth=max_depth, depth=depth + 1)

    # set default timezone to GMT
    set_timezone("GMT")

    initial_data = {}
    cb = fake_cb_response_api(monkeypatch)

    with open(f"{HOME_PATH}/test_data/00007c6f-0000-0a28-01d6-ffde20451832.json", "r") as fp:
        initial_data = json.load(fp)
    proc = Process(cb, "00007c6f-0000-0a28-01d6-ffde20451832", initial_data=initial_data)

    # load events
    for segment_id in initial_data["captured_segments"].keys():
        if segment_id not in proc._events:
            proc._events[segment_id] = {}
        for event_type in ["filemods", "netconns", "regmods", "modloads", "crossprocs", "children"]:
            proc._events[segment_id][event_type] = initial_data["captured_segments"][segment_id][event_type]

    monkeypatch.setattr(proc, "get_segments", _get_segments)
    monkeypatch.setattr(proc, "refresh", _refresh)
    monkeypatch.setattr(proc, "require_events", _require_events)
    monkeypatch.setattr(proc, "_retrieve_cb_info", _retrieve_cb_info)
    monkeypatch.setattr(proc, "all_filemods", _all_filemods)
    monkeypatch.setattr(proc, "all_netconns", _all_netconns)
    monkeypatch.setattr(proc, "all_regmods", _all_regmods)
    monkeypatch.setattr(proc, "all_modloads", _all_modloads)
    monkeypatch.setattr(proc, "all_crossprocs", _all_crossprocs)
    monkeypatch.setattr(proc, "all_childprocs", _all_childprocs)
    monkeypatch.setattr(proc, "walk_children", _walk_children)
    monkeypatch.setattr(proc, "walk_parents", _walk_parents)
    return proc


def test_fake_cb(monkeypatch):
    assert isinstance(fake_cb_response_api(monkeypatch), CbResponseAPI)


def test_make_cb_response_query(monkeypatch):
    from datetime import datetime
    from cbapi.response.models import ProcessQuery
    from cbinterface.response.query import make_process_query

    def _get_query_results(url, query_parameters=None, default=None):
        assert url == "/api/v1/process"
        assert query_parameters == [("cb.urlver", 1), ("q", "hostname:hippo"), ("start", 0), ("rows", 0)]
        return {}

    cb = fake_cb_response_api(monkeypatch)
    result = make_process_query(cb, "hostname:hippo", raise_exceptions=False)
    assert isinstance(result, ProcessQuery)
    result = make_process_query(cb, "hostname:hippo", datetime.now(), raise_exceptions=False)
    assert isinstance(result, ProcessQuery)
    result = make_process_query(cb, "hostname:hippo", datetime.now(), datetime.now(), raise_exceptions=False)
    assert isinstance(result, ProcessQuery)
    monkeypatch.setattr(cb, "get_object", _get_query_results)
    assert len(result) == 0


def test_print_facet_histogram():
    from io import StringIO
    from contextlib import redirect_stdout
    from cbinterface.response.query import print_facet_histogram

    facet_data = {}
    with open(f"{HOME_PATH}/test_data/facet_data.json", "r") as fp:
        facet_data = json.load(fp)
    output_string = ""
    with open(f"{HOME_PATH}/test_data/facet_data_output.txt", "r") as fp:
        output_string = fp.read()
    results = StringIO()
    with redirect_stdout(results):
        print_facet_histogram(facet_data)
    assert results.getvalue() == output_string


def test_make_sensor_query(monkeypatch):
    from cbapi.response import Sensor
    from cbapi.response.models import SensorQuery
    from cbinterface.response.sensor import make_sensor_query

    def _get_sensor_results(url, query_parameters=None, default=None):
        assert url == "/api/v1/sensor"
        assert query_parameters == [("hostname", "hippo")]
        return []

    cb = fake_cb_response_api(monkeypatch)
    monkeypatch.setattr(cb, "get_object", _get_sensor_results)
    result = make_sensor_query(cb, "hostname:hippo")
    assert isinstance(result, SensorQuery)


def test_print_process_info(monkeypatch):
    from cbinterface.response.process import print_process_info

    proc = get_process(monkeypatch)

    result = print_process_info(proc, return_string=True)
    assert isinstance(result, str)
    stored_result = ""
    with open(f"{HOME_PATH}/test_data/printed_process_info.txt", "r") as fp:
        stored_result = fp.read()
    assert result == stored_result
    result = print_process_info(proc, return_string=True, header=False)
    assert "------ INFO ------" not in result


def test_print_process_ancestry(monkeypatch):
    from cbinterface.response.process import print_ancestry

    proc = get_process(monkeypatch)

    expected_output = ""
    with open(f"{HOME_PATH}/test_data/ancestry_str.txt", "r") as fp:
        expected_output = fp.read()
    results = StringIO()
    with redirect_stdout(results):
        print_ancestry(proc)
    assert results.getvalue() == expected_output


def test_print_process_tree(monkeypatch):
    from cbinterface.response.process import print_process_tree

    proc = get_process(monkeypatch)

    expected_output = ""
    with open(f"{HOME_PATH}/test_data/process_tree_str.txt", "r") as fp:
        expected_output = fp.read()
    results = StringIO()
    with redirect_stdout(results):
        print_process_tree(proc)
    assert results.getvalue() == expected_output


def test_print_filemods(monkeypatch):
    from cbinterface.response.process import print_filemods

    proc = get_process(monkeypatch)

    expected_output = ""
    with open(f"{HOME_PATH}/test_data/filemods.txt", "r") as fp:
        expected_output = fp.read()
    results = StringIO()
    with redirect_stdout(results):
        print_filemods(proc)
    assert results.getvalue() == expected_output


def test_print_netconns(monkeypatch):
    from cbinterface.response.process import print_netconns

    proc = get_process(monkeypatch)

    expected_output = ""
    with open(f"{HOME_PATH}/test_data/netconns.txt", "r") as fp:
        expected_output = fp.read()
    results = StringIO()
    with redirect_stdout(results):
        print_netconns(proc)
    assert results.getvalue() == expected_output


def test_print_regmods(monkeypatch):
    from cbinterface.response.process import print_regmods

    proc = get_process(monkeypatch)

    expected_output = ""
    with open(f"{HOME_PATH}/test_data/regmods.txt", "r") as fp:
        expected_output = fp.read()
    results = StringIO()
    with redirect_stdout(results):
        print_regmods(proc)
    assert results.getvalue() == expected_output


def test_print_modloads(monkeypatch):
    from cbinterface.response.process import print_modloads

    proc = get_process(monkeypatch)

    expected_output = ""
    with open(f"{HOME_PATH}/test_data/modloads.txt", "r") as fp:
        expected_output = fp.read()
    results = StringIO()
    with redirect_stdout(results):
        print_modloads(proc)
    assert results.getvalue() == expected_output


def test_print_crossprocs(monkeypatch):
    from cbinterface.response.process import print_crossprocs

    proc = get_process(monkeypatch)

    expected_output = ""
    with open(f"{HOME_PATH}/test_data/crossprocs.txt", "r") as fp:
        expected_output = fp.read()
    results = StringIO()
    with redirect_stdout(results):
        print_crossprocs(proc)
    assert results.getvalue() == expected_output


def test_print_children(monkeypatch):
    from cbinterface.response.process import print_childprocs

    proc = get_process(monkeypatch)

    expected_output = ""
    with open(f"{HOME_PATH}/test_data/children.txt", "r") as fp:
        expected_output = fp.read()
    results = StringIO()
    with redirect_stdout(results):
        print_childprocs(proc)
    assert results.getvalue() == expected_output


def test_process_to_dict(monkeypatch):
    from cbinterface.response.process import process_to_dict

    proc = get_process(monkeypatch)
    cb = proc._cb

    results = process_to_dict(proc, max_segments=2)
    assert isinstance(results, dict)


def test_configured_timezone(monkeypatch):
    from cbinterface.helpers import as_configured_timezone

    proc = get_process(monkeypatch)
    set_timezone("GMT")
    assert "2021-02-10 18:54:14.323000+0000" == as_configured_timezone(proc.start)
    set_timezone("US/Eastern")
    assert "2021-02-10 13:54:14.323000-0500" == as_configured_timezone(proc.start)


"""time zones change time zones
def test_utc_offset_to_potential_tz_names():
    from datetime import timedelta
    from cbinterface.helpers import utc_offset_to_potential_tz_names

    zones = utc_offset_to_potential_tz_names(timedelta(hours=5, minutes=30))
    assert len(zones) == 3
    zones = utc_offset_to_potential_tz_names(timedelta(hours=-5))
    # daylight saving time
    assert len(zones) == 47 or len(zones) == 35
"""



"""Functionality that supports live response "scripts".

In IDR, people like to call things like this playbooks. So, playbooks,
and also remediation scripts to allow for mass remediations.

TODO: Implemented collection scripts for mass collections.
"""

import os
import logging

from configparser import ConfigParser
from typing import Union, List

from cbapi.psc import Device
from cbapi.psc.devices_query import DeviceSearchQuery
from cbapi.live_response_api import CbLRManagerBase

from cbinterface.helpers import get_os_independent_filepath, input_with_timeout

# XXX Make playbooks support all commands?

from cbinterface.commands import (
    PutFile,  # Playbook
    ProcessListing,
    GetFile,  # Playbook
    ListRegKeyValues,
    RegKeyValue,
    ExecuteCommand,  # Playbook
    ListDirectory,
    WalkDirectory,
    LogicalDrives,
    DeleteFile,
    KillProcessByID,
    KillProcessByName,
    DeleteRegistryKeyValue,
    DeleteRegistryKey,
    SetRegKeyValue,
    CreateRegKey,
    GetSystemMemoryDump,
)

# GLOBALS #
LOGGER = logging.getLogger("cbinterface.scripted_live_response")

IGNORED_SECTIONS = ["overview"]

REQUIRED_CMD_KEYS = ["operation"]
REQUIRED_OP_KEY_MAP = {
    "RUN": ["command"],
    "UPLOAD": ["path"],
    "DOWNLOAD": ["file_path", "client_file_path"],
}
""" # remove?
OPTIONAL_CMD_KEYS = ['wait_for_completion', 'get_results']
OPTIONAL_OP_KEY_MAP = {'RUN': ['async_run', 'write_results_path', 'print_results',
                               'common_setup_command', 'common_cleanup_command'],
                       'UPLOAD': ['write_results_path'],
                       'DOWNLOAD': ['client_file_path', 'common_setup_command']
                       } 
"""

# live response scripts - order is important, ordered impacts execution.
SUPPORTED_LR_SCRIPT_KEYS = [
    "pids",
    "process_names",
    "services",
    "scheduled_tasks",
    "registry_values",
    "registry_keys",
    "files",
    "directories",
]

LR_REMEDIATION_MAP = {
    "pids": KillProcessByID,
    "process_names": KillProcessByName,
    "services": "playbook_configs/delete_service.ini",
    "scheduled_tasks": "playbook_configs/delete_scheduled_task.ini",
    "registry_values": DeleteRegistryKeyValue,
    "registry_keys": DeleteRegistryKey,
    "files": DeleteFile,
    "directories": "playbook_configs/delete_directory.ini",
}

# Get the working directory
BASE_DIR = os.path.dirname(os.path.abspath(__file__))

# Playbook section begins #

PLAYBOOK_TEMPLATE_PATH = os.path.join(BASE_DIR, "templates/playbook.ini")


def write_playbook_template(template_path=PLAYBOOK_TEMPLATE_PATH):
    """Write the example template to the current working dir."""
    from shutil import copyfile

    destination = template_path[template_path.rfind("/") + 1 :]
    copyfile(template_path, destination)
    return destination


def playbook_missing_required_keys(config, KEYS):
    """Return True if a playbook is missing required keys."""
    for key in KEYS:
        for section in config.sections():
            if section in IGNORED_SECTIONS:
                return False
            if not config.has_option(section, key):
                LOGGER.error("{} is missing required key: {}".format(section, key))
                return True
    return False


def operation_missing_required_keys(config, section, KEYS):
    """Return True if an operation is missing required keys."""
    for key in KEYS:
        if not config.has_option(section, key):
            LOGGER.error(
                "{} is missing required operation key:{} for operation:{}".format(
                    section, key, config[section]["operation"]
                )
            )
            return True
    return False


def load_playbook(playbook_path):
    """Load playbook config from path."""
    playbook = ConfigParser()
    if not os.path.exists(playbook_path):
        playbook_path = os.path.join(BASE_DIR, playbook_path)
    if not os.path.exists(playbook_path):
        playbook_name = playbook_path[playbook_path.rfind("/") + 1 : playbook_path.rfind(".")]
        LOGGER.error(f"Path to '{playbook_name}' playbook does not exist: {playbook_path}")
        return False
    LOGGER.debug(f"loading playbook path: {playbook_path}")
    try:
        playbook.read(playbook_path)
    except Exception as e:
        LOGGER.error("ConfigParser Error reading '{playbook_path}' : {e}")
        return False
    if playbook_missing_required_keys(playbook, REQUIRED_CMD_KEYS):
        return False
    return playbook


def enforce_argument_placeholders(playbook: ConfigParser, placeholders: dict = {}):
    """Some playbooks require arguments to execute. These arguments are in the form of string format placeholder keys.
    Make sure we have all of the arguments for this playbook. Prompt the user for those arguments if we don't.

    Args:
      playbook: A loaded LR playbook playbook
      placeholders: A dictionary of placeholders we already have (if any)
    Return:
      completed placeholders dict
    """
    LOGGER.debug("Enforcing argument placeholders")
    args_needed = []
    required_args = (
        playbook["overview"]["required_arguments"].split(",")
        if playbook.has_option("overview", "required_arguments")
        else None
    )
    if required_args:
        args_needed = [arg for arg in required_args if arg not in placeholders.keys()]
    for arg in args_needed:
        prompt = f"playbook requires argument {arg}: "
        placeholders[arg] = input_with_timeout(prompt)
    return placeholders


def build_playbook_commands(playbook_path, placeholders={}, separate_cleanup=False):
    """Build the live response commands that define this playbook.

    Args:
        lrsm: A CbLRManagerBase based object (CustomLiveResponseSessionManager)
        devices: iterable of Devices/Sensors (TODO: sensors)
    Returns: List of Cb live response commands.
    """

    # TODO remove cleanup command stuff

    ready_live_response_commands = []

    playbook = load_playbook(playbook_path)
    if not playbook:
        return ready_live_response_commands
    placeholders = enforce_argument_placeholders(playbook, placeholders)

    playbook_name = playbook_path[playbook_path.rfind("/") + 1 : playbook_path.rfind(".")]
    cleanup_commands = []
    playbook_commands = [cmd for cmd in playbook.sections() if cmd not in IGNORED_SECTIONS]
    if separate_cleanup:
        playbook_commands = [
            cmd for cmd in playbook.sections() if cmd not in IGNORED_SECTIONS and not cmd.startswith("cleanup")
        ]
        cleanup_commands = [
            cmd for cmd in playbook.sections() if cmd not in IGNORED_SECTIONS and cmd.startswith("cleanup")
        ]

    # make sure requirements are met first
    for command in playbook_commands:
        op = playbook[command]["operation"].upper()
        if op not in REQUIRED_OP_KEY_MAP:
            LOGGER.error("{op} is not a recognized operation")
            return False
        if operation_missing_required_keys(playbook, command, REQUIRED_OP_KEY_MAP[op]):
            return False

    LOGGER.info(f"building live response commands defined by {playbook_name}")

    for command in playbook_commands:
        op = playbook[command]["operation"].upper()
        post_completion_command = playbook[command].get("post_completion_command", None)

        if op == "RUN":
            command_string = playbook[command]["command"]
            wait_for_output = playbook[command].getboolean("wait_for_output", True)
            remote_output_file_name = playbook[command].get("remote_output_file_name", None)
            working_directory = playbook[command].get("working_directory", None)
            wait_timeout = playbook[command].getint("wait_timeout", 30)
            wait_for_completion = playbook[command].getboolean("wait_for_completion", True)
            print_results = playbook[command].getboolean("print_results", True)
            write_results_path = playbook[command].get("write_results_path", False)

            cmd = ExecuteCommand(
                command_string,
                wait_for_output=wait_for_output,
                remote_output_file_name=remote_output_file_name,
                working_directory=working_directory,
                wait_timeout=wait_timeout,
                wait_for_completion=wait_for_completion,
                print_results=print_results,
                write_results_path=write_results_path,
                placeholders=placeholders,
                post_completion_command=post_completion_command,
            )
            LOGGER.debug(f"built {cmd}")
            ready_live_response_commands.append(cmd)

        elif op == "DOWNLOAD" or op == "PUT":
            file_path = playbook[command]["file_path"]
            client_file_path = playbook[command]["client_file_path"]

            if not os.path.exists(file_path):
                original_fp = file_path
                file_path = os.path.join(BASE_DIR, file_path)
                if not os.path.exists(file_path):
                    LOGGER.error(f"Not found: '{original_fp}' OR '{file_path}'")
                    return False

            file_name = get_os_independent_filepath(file_path).name

            cmd = PutFile(
                file_path,
                sensor_write_filepath=client_file_path,
                placeholders=placeholders,
                post_completion_command=post_completion_command,
            )
            cmd.description = f"Put '{file_name}' on device @ '{client_file_path}'"
            LOGGER.debug(f"built {cmd}")
            ready_live_response_commands.append(cmd)

        elif op == "UPLOAD" or op == "GET":
            path = playbook[command]["path"]
            write_results_path = playbook[command].get("write_results_path", None)
            cmd = GetFile(
                path,
                output_filename=write_results_path,
                placeholders=placeholders,
                post_completion_command=post_completion_command,
            )
            LOGGER.debug(f"built {cmd}")
            ready_live_response_commands.append(cmd)

    if separate_cleanup:
        ready_live_response_cleanup_commands = []
        unique_cleanup_commands = []
        for command in cleanup_commands:
            op = playbook[command]["operation"].upper()

            if op == "RUN":
                command_string = playbook[command]["command"]
                if command_string in unique_cleanup_commands:
                    continue
                unique_cleanup_commands.append(command_string)
                wait_for_output = playbook[command].getboolean("wait_for_output", True)
                remote_output_file_name = playbook[command].get("remote_output_file_name", None)
                working_directory = playbook[command].getboolean("working_directory", None)
                wait_timeout = playbook[command].getint("wait_timeout", 30)
                wait_for_completion = playbook[command].getboolean("wait_for_completion", True)
                print_results = playbook[command].getboolean("print_results", True)
                write_results_path = playbook[command].get("write_results_path", False)

                cmd = ExecuteCommand(
                    command_string,
                    wait_for_output=wait_for_output,
                    remote_output_file_name=remote_output_file_name,
                    working_directory=working_directory,
                    wait_timeout=wait_timeout,
                    wait_for_completion=wait_for_completion,
                    print_results=print_results,
                    write_results_path=write_results_path,
                    placeholders=placeholders,
                )
                LOGGER.debug(f"built {cmd}")
                ready_live_response_commands.append(cmd)

        return ready_live_response_commands, ready_live_response_cleanup_commands

    return ready_live_response_commands


# Remediation section begins #

REMEDIATION_TEMPLATE_PATH = os.path.join(BASE_DIR, "templates/remediate.ini")


def write_remediation_template(template_path=REMEDIATION_TEMPLATE_PATH):
    """Write the example template to the current working dir."""
    from shutil import copyfile

    destination = template_path[template_path.rfind("/") + 1 :]
    copyfile(template_path, destination)
    return destination


def load_live_response_script(remediation_script_path):
    """Load live response script config from path."""
    script = ConfigParser()
    if not os.path.exists(remediation_script_path):
        remediation_script_path = os.path.join(BASE_DIR, remediation_script_path)
    if not os.path.exists(remediation_script_path):
        script_name = remediation_script_path[
            remediation_script_path.rfind("/") + 1 : remediation_script_path.rfind(".")
        ]
        LOGGER.error(f"Path to '{script_name}' script does not exist: {remediation_script_path}")
        return False
    LOGGER.debug(f"loading script path: {remediation_script_path}")
    try:
        script.read(remediation_script_path)
    except Exception as e:
        LOGGER.error("ConfigParser Error reading '{remediation_script_path}' : {e}")
        return False
    return script


def build_remediation_commands(remediation_script_path):
    """Return Live Response Session Commands described by this script."""
    from cbinterface.config import get_playbook_map

    file_paths = []
    process_names = []
    process_ids = []
    scheduled_tasks = []
    services = []
    directories = []
    reg_key_values = []
    reg_keys = []

    script = load_live_response_script(remediation_script_path)
    if not script:
        return False

    cleanup_commands = []
    remediation_commands = []

    for remediation_type in SUPPORTED_LR_SCRIPT_KEYS:
        remediation_function_or_play = LR_REMEDIATION_MAP[remediation_type]

        if script.has_section(remediation_type):
            for remediation_key, remediation_item in script[remediation_type].items():
                if remediation_function_or_play is NotImplemented:
                    LOGGER.warning(f"{remediation_type} remediation is NotImplemented yet.")
                    continue
                elif isinstance(remediation_function_or_play, str) and remediation_function_or_play.startswith(
                    "playbook"
                ):
                    playbook = load_playbook(remediation_function_or_play)
                    if not playbook:
                        continue
                    placeholder_argument_name = playbook.get("overview", "required_arguments", fallback="").split(",")[
                        0
                    ]
                    if not placeholder_argument_name:
                        continue
                    placeholders = {placeholder_argument_name: remediation_item}
                    cmds = build_playbook_commands(remediation_function_or_play, placeholders=placeholders)
                    remediation_commands.extend(cmds)
                else:
                    cmd = remediation_function_or_play(remediation_item)
                    LOGGER.info(f"created {cmd}")
                    remediation_commands.append(cmd)

    return remediation_commands



"""Everything Watchlist management.

I wrote everything here to help migrate watchlists to to PSC EDR.
"""

import datetime
import logging
from dateutil import tz

from typing import Union, List, Dict

from cbapi.query import SimpleQuery
from cbapi.response import CbResponseAPI, Watchlist

LOGGER = logging.getLogger("cbinterface.response.watchlists")


def get_all_watchlists(cb: CbResponseAPI) -> SimpleQuery:
    """Return a list of all watchlists."""
    return cb.select(Watchlist)


def query_watchlists(cb: CbResponseAPI, query: str) -> SimpleQuery:
    """perform watchlist query"""
    try:
        return cb.select(Watchlist).where(query)
    except Exception as e:
        LOGGER.error(f"problem querying watchlists: {e}")
        return []


def watchlist_to_dict(watchlist: Watchlist) -> Dict:
    """Return watchlist dictionary."""
    wl = watchlist._info
    wl["query"] = watchlist.query
    return wl


def these_watchlists_to_list_dict(cb: CbResponseAPI, watchlist_names=[], watchlist_ids=[]) -> List[Dict]:
    """Convert the listed watchlists to a list of their dictionary representations."""
    wl_data = []
    for wl_name in watchlist_names:
        wl = cb.select(Watchlist).where(f"name:{wl_name}")
        if wl:
            if len(wl) > 1:
                LOGGER.warning(f"got {len(wl)} watchlists with name matching {wl_name}. Using first result")
            wl = wl[0]
            wl_data.append(watchlist_to_dict(wl))

    return wl_data
